# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime as dt
import os

import mock

from rally.env import env_mgr
from rally.env import platform
from rally import exceptions
from tests.unit import test


class EnvManagerTestCase(test.TestCase):

    def test_init(self):
        data = {"uuid": "any", "balbalba": "balbal"}
        mgr = env_mgr.EnvManager(data)
        self.assertEqual("any", mgr.uuid)
        expected = copy.deepcopy(data)
        expected["platforms"] = []
        self.assertEqual(expected, mgr._env)

    def test__str__(self):
        data = {"uuid": "env_uuid", "name": "env_name"}
        mgr = env_mgr.EnvManager(data)
        self.assertEqual("Env `env_name (env_uuid)'", str(mgr))

    @mock.patch("rally.common.db.env_get_status")
    def test_status_property(self, mock_env_get_status):
        self.assertEqual(mock_env_get_status.return_value,
                         env_mgr.EnvManager({"uuid": "any"}).status)

        mock_env_get_status.assert_called_once_with("any")

    @mock.patch("rally.common.db.platforms_list")
    @mock.patch("rally.common.db.env_get")
    def test_data_property(self, mock_env_get, mock_platforms_list):
        created_at = dt.datetime.now()
        updated_at = dt.datetime.now()
        mock_platforms_list.return_value = [
            {"created_at": created_at, "updated_at": updated_at,
             "platform_name": "foo"}]

        mock_env_get.return_value = {
            "id": "66",
            "uuid": "666",
            "created_at": created_at,
            "updated_at": updated_at,
            "name": "42",
            "description": "Some description",
            "status": "some status",
            "spec": "some_spec",
            "extras": "some_extras",
        }

        result = env_mgr.EnvManager({"uuid": 111}).data
        for key in ["name", "description", "status", "spec", "extras", "uuid"]:
            self.assertEqual(mock_env_get.return_value[key], result[key])
        self.assertEqual(created_at.isoformat(), result["created_at"])
        self.assertEqual(updated_at.isoformat(), result["updated_at"])
        self.assertEqual(
            {"foo": {"created_at": created_at.isoformat(),
                     "updated_at": updated_at.isoformat(),
                     "platform_name": "foo"}},
            result["platforms"]
        )
        mock_platforms_list.assert_called_once_with(111)
        mock_env_get.assert_called_once_with(111)

    @mock.patch("rally.common.db.platforms_list")
    def test__get_platforms(self, mock_platforms_list):

        @platform.configure(name="some", platform="foo")
        class FooPlatform(platform.Platform):
            pass

        mock_platforms_list.side_effect = [
            [],
            [
                {
                    "uuid": "1",
                    "plugin_name": "some@foo",
                    "plugin_data": "plugin_data",
                    "plugin_spec": "plugin_data",
                    "platform_data": "platform_data",
                    "status": "INIT"
                },
                {
                    "uuid": "2",
                    "plugin_name": "some@foo",
                    "plugin_data": None,
                    "plugin_spec": "plugin_data",
                    "platform_data": None,
                    "status": "CREATED"
                },
            ]
        ]

        self.assertEqual([], env_mgr.EnvManager({"uuid": 42})._get_platforms())
        mock_platforms_list.assert_called_once_with(42)

        result = env_mgr.EnvManager({"uuid": 43})._get_platforms()
        self.assertEqual(2, len(result))

        for i, r in enumerate(sorted(result, key=lambda x: x.uuid)):
            self.assertIsInstance(r, FooPlatform)
            self.assertEqual("some@foo", r.get_fullname())
            self.assertEqual(str(i + 1), r.uuid)

        mock_platforms_list.assert_has_calls([mock.call(42), mock.call(43)])

    @mock.patch("rally.common.db.env_get",
                return_value={"uuid": "1"})
    def test_get(self, mock_env_get):
        self.assertEqual("1", env_mgr.EnvManager.get("1").uuid)
        mock_env_get.assert_called_once_with("1")

    @mock.patch("rally.common.db.env_list",
                return_value=[{"uuid": "1"}, {"uuid": "2"}])
    def test_list(self, mock_env_list):
        result = env_mgr.EnvManager.list()

        for r in result:
            self.assertIsInstance(r, env_mgr.EnvManager)

        self.assertEqual(set(r.uuid for r in result), set(["1", "2"]))
        self.assertEqual(2, len(result))
        mock_env_list.assert_called_once_with(status=None)

    @mock.patch("rally.common.db.env_create")
    def test__validate_and_create_env_empty_spec(self, mock_env_create):
        mock_env_create.return_value = {"uuid": "1"}
        env = env_mgr.EnvManager._validate_and_create_env("name", {})

        self.assertIsInstance(env, env_mgr.EnvManager)
        self.assertEqual("1", env.uuid)
        mock_env_create.assert_called_once_with(
            "name", env_mgr.STATUS.INIT, "", {}, {}, {}, [])

    def test__validate_and_create_env_invalid_spec(self):
        self.assertRaises(
            exceptions.ManagerInvalidSpec,
            env_mgr.EnvManager._validate_and_create_env,
            "n", {"!description": {"it's": "wrong"}}
        )

    @mock.patch("rally.common.db.env_create")
    def test__validate_and_create_env_with_spec(self, mock_env_create):
        mock_env_create.return_value = {"uuid": "1"}

        @platform.configure("existing", platform="valid1")
        class Platform1(platform.Platform):
            CONFIG_SCHEMA = {
                "type": "object",
                "properties": {"a": {"type": "string"}},
                "additionalProperties": False
            }

        @platform.configure("other", platform="valid1")
        class Platform1b(Platform1):
            pass

        @platform.configure("2", platform="valid2")
        class Platform2(platform.Platform):
            CONFIG_SCHEMA = {
                "type": "object",
                "properties": {"b": {"type": "string"}},
                "additionalProperties": False
            }

        self.addCleanup(Platform1.unregister)
        self.addCleanup(Platform2.unregister)

        env_mgr.EnvManager._validate_and_create_env(
            "n",
            {
                "!description": "d",
                "!extras": {"ext": 1},
                "valid1": {"a": "str"}
            }
        )

        expected_platforms = [{
            "status": platform.STATUS.INIT,
            "plugin_name": "existing@valid1",
            "plugin_spec": {"a": "str"},
            "platform_name": "valid1"
        }]

        mock_env_create.assert_called_once_with(
            "n", env_mgr.STATUS.INIT, "d", {"ext": 1}, {},
            {"existing@valid1": {"a": "str"}}, expected_platforms)

        mock_env_create.reset_mock()
        self.assertRaises(
            exceptions.ManagerInvalidSpec,
            env_mgr.EnvManager._validate_and_create_env,
            "n", {"valid1": {"a": "str"}, "2@valid2": {"c": 1}})
        self.assertFalse(mock_env_create.called)

        mock_env_create.reset_mock()
        self.assertRaises(
            exceptions.ManagerInvalidSpec,
            env_mgr.EnvManager._validate_and_create_env,
            "n", {"non_existing@nope": {"a": "str"}}
        )
        self.assertFalse(mock_env_create.called)

        self.assertRaises(
            exceptions.ManagerInvalidSpec,
            env_mgr.EnvManager._validate_and_create_env,
            "n", {"valid1": {"a": "str"}, "other@valid1": {"a": "str"}}
        )

    @mock.patch("rally.common.db.env_set_status")
    @mock.patch("rally.common.db.platform_set_data")
    @mock.patch("rally.common.db.platform_set_status")
    @mock.patch("rally.common.db.platforms_list")
    def test__create_platforms(self,
                               mock_platforms_list, mock_platform_set_status,
                               mock_platform_set_data, mock_env_set_status):

        # One platform that passes successfully
        @platform.configure("passes", platform="create")
        class ValidPlatform(platform.Platform):
            def create(self):
                return {"platform": "data"}, {"plugin": "data"}

        self.addCleanup(ValidPlatform.unregister)
        mock_platforms_list.return_value = [
            {
                "uuid": "p_uuid",
                "plugin_name": "passes@create",
                "plugin_spec": {},
                "plugin_data": None,
                "platform_data": None,
                "status": platform.STATUS.INIT
            }
        ]
        env_mgr.EnvManager({"uuid": 121})._create_platforms()

        mock_platforms_list.assert_called_once_with(121)
        mock_platform_set_status.assert_called_once_with(
            "p_uuid", platform.STATUS.INIT, platform.STATUS.READY)
        mock_platform_set_data.assert_called_once_with(
            "p_uuid",
            platform_data={"platform": "data"}, plugin_data={"plugin": "data"})
        mock_env_set_status.assert_called_once_with(
            121, env_mgr.STATUS.INIT, env_mgr.STATUS.READY)

    @mock.patch("rally.common.db.env_set_status")
    @mock.patch("rally.common.db.platform_set_status")
    @mock.patch("rally.env.env_mgr.EnvManager._get_platforms")
    def test__create_platforms_failed(self, mock__get_platforms,
                                      mock_platform_set_status,
                                      mock_env_set_status):
        # Check when first fails, second is marked as skipped

        @platform.configure("bad", platform="create")
        class InValidPlatform(platform.Platform):
            def create(self):
                raise Exception("I don't want to work!")

        @platform.configure("good_but_skipped", platform="create")
        class ValidPlatform(platform.Platform):
            def create(self):
                return {"platform": "data"}, {"plugin": "data"}

        for p in [InValidPlatform, ValidPlatform]:
            self.addCleanup(p.unregister)

        mock__get_platforms.return_value = [
            InValidPlatform({}, uuid=1), ValidPlatform({}, uuid=2)]

        env_mgr.EnvManager({"uuid": 42})._create_platforms()

        mock_env_set_status.assert_called_once_with(
            42, env_mgr.STATUS.INIT, env_mgr.STATUS.FAILED_TO_CREATE)

        mock_platform_set_status.assert_has_calls([
            mock.call(1, platform.STATUS.INIT,
                      platform.STATUS.FAILED_TO_CREATE),
            mock.call(2, platform.STATUS.INIT, platform.STATUS.SKIPPED),
        ])

    @mock.patch("rally.common.db.env_set_status")
    @mock.patch("rally.common.db.platform_set_status")
    @mock.patch("rally.common.db.platform_set_data")
    @mock.patch("rally.env.env_mgr.EnvManager._get_platforms")
    def test__create_platforms_when_db_issues_autodestroy(
            self, mock__get_platforms, mock_platform_set_data,
            mock_platform_set_status, mock_env_set_status):
        # inject db errors check that auto destroy is called
        platform1 = mock.MagicMock()
        platform1.uuid = 11
        platform1.destroy.side_effect = Exception
        platform1.create.return_value = ("platform_d", "plugin_d")
        platform2 = mock.MagicMock()
        platform2.uuid = 22
        mock__get_platforms.return_value = [platform1, platform2]
        mock_platform_set_data.side_effect = Exception

        env_mgr.EnvManager({"uuid": 42})._create_platforms()

        mock_platform_set_status.assert_called_once_with(
            22, platform.STATUS.INIT, platform.STATUS.SKIPPED)
        mock_platform_set_data.assert_called_once_with(
            11, platform_data="platform_d", plugin_data="plugin_d")
        mock_env_set_status.assert_called_once_with(
            42, env_mgr.STATUS.INIT, env_mgr.STATUS.FAILED_TO_CREATE)

    @mock.patch("rally.common.db.platforms_list", return_value=[])
    @mock.patch("rally.common.db.env_set_status")
    @mock.patch("rally.common.db.env_create", return_value={"uuid": 121})
    def test_create(self, mock_env_create, mock_env_set_status,
                    mock_platforms_list):
        # NOTE(boris-42): Just check with empty spec that just check workflow
        result = env_mgr.EnvManager.create("a", {})

        self.assertIsInstance(result, env_mgr.EnvManager)
        self.assertEqual(121, result.uuid)

    @mock.patch("rally.env.env_mgr.EnvManager._create_platforms")
    @mock.patch("rally.env.env_mgr.EnvManager._validate_and_create_env")
    def test_create_override_spec_values(self, mock__validate_and_create_env,
                                         mock__create_platforms):

        default_spec = {
            "!description": "default description",
            "!extras": {"e": 1},
            "!config": {"c": 2},
            "some@platform": {}
        }
        env_mgr.EnvManager.create(
            "a", default_spec, description="d", config="c", extras="e")
        mock__validate_and_create_env.assert_called_once_with(
            "a",
            {
                "!description": "d", "!config": "c", "!extras": "e",
                "some@platform": {}
            }
        )

    @mock.patch.dict(os.environ, values={"KEY": "value"}, clear=True)
    @mock.patch("rally.env.platform.Platform.get_all")
    def test_create_spec_from_sys_environ(self, mock_platform_get_all):

        # let's cover all positive and minor cases at once

        class Foo1Platform(platform.Platform):
            """This platform doesn't override original methods"""

            @classmethod
            def get_fullname(cls):
                return cls.__name__

        class Foo2Platform(Foo1Platform):
            """This platform should try to modify sys environment"""

            @classmethod
            def create_spec_from_sys_environ(cls, sys_environ):
                for key in list(sys_environ.keys()):
                    sys_environ.pop(key)
                return platform.Platform.create_spec_from_sys_environ({})

        class Foo3Platform(Foo1Platform):
            """This platform rely on one sys argument."""

            @classmethod
            def create_spec_from_sys_environ(cls, sys_environ):
                self.assertIn("KEY", sys_environ)
                return {"spec": {"KEY": sys_environ["KEY"]},
                        "available": True}

        class Foo4Platform(Foo1Platform):
            """This platform raises an error!"""

            @classmethod
            def create_spec_from_sys_environ(cls, sys_environ):
                raise KeyError("Ooopes")

        class Foo5Platform(Foo1Platform):
            """This platform returns invalid data."""

            @classmethod
            def create_spec_from_sys_environ(cls, sys_environ):
                return {"foo": "bar"}

        mock_platform_get_all.return_value = [
            Foo1Platform, Foo2Platform, Foo3Platform, Foo4Platform,
            Foo5Platform]

        result = env_mgr.EnvManager.create_spec_from_sys_environ()
        self.assertEqual({"Foo3Platform": {"KEY": "value"}}, result["spec"])
        self.assertEqual(
            {"Foo1Platform", "Foo2Platform", "Foo3Platform", "Foo4Platform",
             "Foo5Platform"},
            set(result["discovery_details"].keys()))
        result = result["discovery_details"]
        default_msg = "Skipped. No credentials found."

        self.assertFalse(result["Foo1Platform"]["available"])
        self.assertEqual(default_msg, result["Foo1Platform"]["message"])

        self.assertFalse(result["Foo2Platform"]["available"])
        self.assertEqual(default_msg, result["Foo2Platform"]["message"])

        self.assertTrue(result["Foo3Platform"]["available"])
        self.assertEqual("Available", result["Foo3Platform"]["message"])

        self.assertFalse(result["Foo4Platform"]["available"])
        self.assertIn("method is broken", result["Foo4Platform"]["message"])
        self.assertIn("traceback", result["Foo4Platform"])

        self.assertFalse(result["Foo5Platform"]["available"])
        self.assertIn("method is broken", result["Foo5Platform"]["message"])
        self.assertNotIn("traceback", result["Foo5Platform"])

        mock_platform_get_all.assert_called_once_with()

    @mock.patch("rally.common.db.env_rename")
    def test_rename(self, mock_env_rename):
        env = env_mgr.EnvManager({"uuid": "11", "name": "n"})

        self.assertTrue(env.rename("n"))
        self.assertEqual(0, mock_env_rename.call_count)
        self.assertTrue(env.rename("n2"))
        mock_env_rename.assert_called_once_with("11", "n", "n2")

    @mock.patch("rally.common.db.env_update")
    def test_update(self, mock_env_update):
        env = env_mgr.EnvManager(
            {"uuid": "11", "description": "d", "extras": "e"})
        env.update(description="d2", config="c2", extras="e2")
        mock_env_update.assert_called_once_with(
            "11", description="d2", config="c2", extras="e2")

    def test_update_spec(self):
        self.assertRaises(NotImplementedError,
                          env_mgr.EnvManager({"uuid": 1}).update_spec, "")

    @mock.patch("rally.env.env_mgr.EnvManager._get_platforms")
    def test_check_health(self, mock__get_platforms):

        valid_result = {
            "available": False,
            "message": "Nope I don't want to work"
        }

        @platform.configure(name="valid", platform="check")
        class ValidPlatform(platform.Platform):

            def check_health(self):
                return valid_result

        @platform.configure(name="broken_fromat", platform="check")
        class BrokenFormatPlatform(platform.Platform):

            def check_health(self):
                return {"something": "is wrong here in format"}

        @platform.configure(name="just_broken", platform="check")
        class JustBrokenPlatform(platform.Platform):

            def check_health(self):
                raise Exception("This is really bad exception")

        for p in [ValidPlatform, BrokenFormatPlatform, JustBrokenPlatform]:
            self.addCleanup(p.unregister)

        mock__get_platforms.side_effect = [
            [ValidPlatform("spec1")],
            [ValidPlatform("spec1"), BrokenFormatPlatform("spec2")],
            [JustBrokenPlatform("spec3")]
        ]

        self.assertEqual({"valid@check": valid_result},
                         env_mgr.EnvManager({"uuid": "42"}).check_health())

        broken_msg = "Plugin %s.check_health() method is broken"

        self.assertEqual(
            {
                "valid@check": valid_result,
                "broken_fromat@check": {
                    "message": broken_msg % "broken_fromat@check",
                    "available": False
                }
            },
            env_mgr.EnvManager({"uuid": "43"}).check_health())

        self.assertEqual(
            {
                "just_broken@check": {
                    "message": broken_msg % "just_broken@check",
                    "available": False,
                    "traceback": mock.ANY
                }
            },
            env_mgr.EnvManager({"uuid": "44"}).check_health())

        mock__get_platforms.assert_has_calls([mock.call()] * 3)

    @mock.patch("rally.env.env_mgr.EnvManager._get_platforms")
    def test_get_info(self, mock__get_platforms):

        @platform.configure(name="valid", platform="info")
        class InfoValid(platform.Platform):

            def info(self):
                return {"info": "it works!", "error": ""}

        @platform.configure(name="wrong_fmt", platform="info")
        class InfoWrongFormat(platform.Platform):

            def info(self):
                return {"something": "is wrong"}

        @platform.configure(name="broken", platform="info")
        class InfoBroken(platform.Platform):

            def info(self):
                raise Exception("This should not happen")

        for p in [InfoValid, InfoWrongFormat, InfoBroken]:
            self.addCleanup(p.unregister)

        mock__get_platforms.side_effect = [
            [InfoValid("spec1")],
            [InfoValid("spec1"), InfoWrongFormat("spec2")],
            [InfoValid("spec1"), InfoBroken("spec3")],
        ]

        valid_result = {"info": "it works!", "error": ""}

        self.assertEqual({"valid@info": valid_result},
                         env_mgr.EnvManager({"uuid": "42"}).get_info())

        self.assertEqual(
            {
                "valid@info": valid_result,
                "wrong_fmt@info": {
                    "error": "Plugin wrong_fmt@info.info() method is broken",
                    "info": None
                }
            },
            env_mgr.EnvManager({"uuid": "43"}).get_info())

        self.assertEqual(
            {
                "valid@info": valid_result,
                "broken@info": {
                    "error": "Plugin broken@info.info() method is broken",
                    "info": None,
                    "traceback": mock.ANY
                }
            },
            env_mgr.EnvManager({"uuid": "44"}).get_info())

        mock__get_platforms.assert_has_calls([mock.call()] * 3)

    @mock.patch("rally.common.db.env_set_status")
    @mock.patch("rally.env.env_mgr.EnvManager._get_platforms")
    def test_cleanup(self, mock__get_platforms, mock_env_set_status):

        valid_result = {
            "discovered": 10,
            "deleted": 6,
            "failed": 4,
            "resources": {
                "vm": {
                    "discovered": 2,
                    "failed": 2,
                    "deleted": 0
                }
            },
            "errors": [
                {
                    "resource_id": "1",
                    "resource_type": "vm",
                    "message": "something"
                }
            ]
        }

        @platform.configure(name="valid", platform="clean")
        class CleanValid(platform.Platform):

            def cleanup(self, task_uuid=None):
                return dict(valid_result)

        @platform.configure(name="wrong", platform="clean")
        class CleanWrongFormat(platform.Platform):

            def cleanup(self, task_uuid):
                return {"something": "is wrong"}

        @platform.configure(name="broken", platform="clean")
        class CleanBroken(platform.Platform):

            def cleanup(self, task_uuid):
                raise Exception("This should not happen")

        @platform.configure(name="not_impl", platform="clean")
        class NotImplBroken(platform.Platform):
            pass

        for p in [CleanValid, CleanWrongFormat, CleanBroken, NotImplBroken]:
            self.addCleanup(p.unregister)

        mock__get_platforms.return_value = [
            CleanValid("spec1"), CleanBroken("spec2"),
            CleanWrongFormat("spec3"), NotImplBroken("spec4")
        ]

        result = env_mgr.EnvManager({"uuid": 424}).cleanup()
        mock__get_platforms.assert_called_once_with()
        mock_env_set_status.assert_has_calls([
            mock.call(424, env_mgr.STATUS.READY, env_mgr.STATUS.CLEANING),
            mock.call(424, env_mgr.STATUS.CLEANING, env_mgr.STATUS.READY)
        ])
        self.assertIsInstance(result, dict)
        self.assertEqual(4, len(result))
        valid_result["message"] = "Succeeded"
        self.assertEqual(valid_result, result["valid@clean"])
        self.assertEqual(
            {
                "message": "Failed",
                "discovered": 0, "deleted": 0, "failed": 0, "resources": {},
                "errors": [{
                    "message": "Plugin wrong@clean.cleanup() method is broken",
                }]
            },
            result["wrong@clean"]
        )
        self.assertEqual(
            {
                "message": "Failed",
                "discovered": 0, "deleted": 0, "failed": 0, "resources": {},
                "errors": [{
                    "message": "Plugin broken@clean.cleanup() method is "
                               "broken",
                    "traceback": mock.ANY
                }]
            },
            result["broken@clean"]
        )
        self.assertEqual(
            {
                "message": "Not implemented",
                "discovered": 0, "deleted": 0, "failed": 0, "resources": {},
                "errors": []
            },
            result["not_impl@clean"]
        )

    @mock.patch("rally.env.env_mgr.EnvManager.cleanup")
    def test_destroy_cleanup_failed(self, mock_env_manager_cleanup):
        mock_env_manager_cleanup.return_value = {
            "platform_1": {
                "errors": [],
            },
            "platform_2": {
                "errors": [121],
            }
        }
        self.assertEqual(
            {
                "cleanup_info": {
                    "skipped": False,
                    "failed": True,
                    "info": {
                        "platform_1": {"errors": []},
                        "platform_2": {"errors": [121]}
                    }
                },
                "destroy_info": {
                    "skipped": True,
                    "platforms": {},
                    "message": "Skipped because cleanup has errors"
                }
            },
            env_mgr.EnvManager({"uuid": 42}).destroy()
        )
        mock_env_manager_cleanup.assert_called_once_with()

    @mock.patch("rally.common.objects.Verifier.list")
    @mock.patch("rally.env.env_mgr.EnvManager._get_platforms", return_value=[])
    @mock.patch("rally.common.db.env_set_status")
    def test_destroy_no_platforms(self, mock_env_set_status,
                                  mock__get_platforms, mock_verifier_list):
        self.assertEqual(
            {
                "cleanup_info": {"skipped": True},
                "destroy_info": {"skipped": False, "platforms": {}}
            },
            env_mgr.EnvManager({"uuid": 42}).destroy(skip_cleanup=True)
        )
        mock_env_set_status.assert_has_calls([
            mock.call(42, env_mgr.STATUS.READY, env_mgr.STATUS.DESTROYING),
            mock.call(42, env_mgr.STATUS.DESTROYING, env_mgr.STATUS.DESTROYED)
        ])
        mock__get_platforms.assert_called_once_with()

    @mock.patch("rally.common.objects.Verifier.list")
    @mock.patch("rally.common.db.env_set_status")
    @mock.patch("rally.common.db.platform_set_status")
    @mock.patch("rally.env.env_mgr.EnvManager._get_platforms")
    def test_destroy_with_platforms(self, mock__get_platforms,
                                    mock_platform_set_status,
                                    mock_env_set_status,
                                    mock_verifier_list):
        verifier = mock.Mock()
        mock_verifier_list.return_value = [verifier]

        platform1 = mock.MagicMock()
        platform1.get_fullname.return_value = "p_destroyed"
        platform1.status = platform.STATUS.DESTROYED

        platform2 = mock.MagicMock()
        platform2.get_fullname.return_value = "p_valid"
        platform2.status = platform.STATUS.READY

        platform3 = mock.MagicMock()
        platform3.get_fullname.return_value = "p_invalid"
        platform3.destroy.side_effect = Exception
        platform3.status = platform.STATUS.READY

        mock__get_platforms.return_value = [
            platform1, platform2, platform3
        ]

        result = env_mgr.EnvManager({"uuid": 666}).destroy(skip_cleanup=True)
        self.assertIsInstance(result, dict)
        self.assertEqual(2, len(result))
        self.assertEqual({"skipped": True}, result["cleanup_info"])
        self.assertEqual(
            {
                "skipped": False,
                "platforms": {
                    "p_destroyed": {
                        "message": "Platform is already destroyed. Do nothing",
                        "status": {
                            "old": platform.STATUS.DESTROYED,
                            "new": platform.STATUS.DESTROYED
                        },
                    },
                    "p_valid": {
                        "message": "Successfully destroyed",
                        "status": {
                            "old": platform.STATUS.READY,
                            "new": platform.STATUS.DESTROYED
                        },
                    },
                    "p_invalid": {
                        "message": "Failed to destroy",
                        "status": {
                            "old": platform.STATUS.READY,
                            "new": platform.STATUS.FAILED_TO_DESTROY
                        },
                        "traceback": mock.ANY
                    }
                }
            },
            result["destroy_info"])

        mock__get_platforms.assert_called_once_with()
        mock_platform_set_status.assert_has_calls([
            mock.call(platform2.uuid,
                      platform.STATUS.READY,
                      platform.STATUS.DESTROYING),
            mock.call(platform2.uuid,
                      platform.STATUS.DESTROYING,
                      platform.STATUS.DESTROYED),
            mock.call(platform3.uuid,
                      platform.STATUS.READY,
                      platform.STATUS.DESTROYING),
            mock.call(platform3.uuid,
                      platform.STATUS.DESTROYING,
                      platform.STATUS.FAILED_TO_DESTROY)
        ])

        verifier.set_env.assert_called_once_with(666)
        verifier.manager.uninstall.assert_called_once_with()

    @mock.patch("rally.common.db.env_get_status")
    @mock.patch("rally.common.db.env_delete_cascade")
    def test_delete(self, mock_env_delete_cascade, mock_env_get_status):
        mock_env_get_status.side_effect = [
            "WRONG", env_mgr.STATUS.DESTROYED
        ]
        self.assertRaises(exceptions.ManagerInvalidState,
                          env_mgr.EnvManager({"uuid": "42"}).delete)
        self.assertFalse(mock_env_delete_cascade.called)

        env_mgr.EnvManager({"uuid": "43"}).delete()
        mock_env_delete_cascade.assert_called_once_with("43")

        mock_env_get_status.assert_has_calls(
            [mock.call("42"), mock.call("43")])

    @mock.patch("rally.common.db.env_get_status")
    @mock.patch("rally.common.db.env_delete_cascade")
    def test_delete_force(self, mock_env_delete_cascade, mock_env_get_status):
        mock_env_get_status.return_value = "WRONG"
        env_mgr.EnvManager({"uuid": "44"}).delete(force=True)
        mock_env_delete_cascade.assert_called_once_with("44")

    @mock.patch("rally.env.env_mgr.EnvManager._get_platforms")
    def test_get_validation_context(self, mock__get_platforms):
        platform1 = mock.MagicMock()
        platform1._get_validation_context.return_value = {
            "users@openstack": {}}

        platform2 = mock.MagicMock()
        platform2._get_validation_context.return_value = {
            "foo_bar": "xxx"}

        mock__get_platforms.return_value = [platform1, platform2]

        env = env_mgr.EnvManager({"uuid": "44"})

        self.assertEqual({"users@openstack": {},
                          "foo_bar": "xxx"}, env.get_validation_context())
        platform1._get_validation_context.assert_called_once_with()
        platform2._get_validation_context.assert_called_once_with()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import testtools

from tests.functional import utils


class DeploymentTestCase(testtools.TestCase):

    def test_create_deployment_from_env(self):
        os.environ.update(
            {
                "OS_AUTH_URL": "http://fake",
                "OS_USERNAME": "fake",
                "OS_PASSWORD": "fake"
            }
        )
        rally = utils.Rally(plugin_path="tests/functional/extra")
        rally("deployment create --name fromenv --fromenv")
        spec = rally("deployment config", getjson=True)
        self.assertEqual({
            "good@fake": {
                "auth_url": "http://fake",
                "username": "fake",
                "password": "fake"
            }
        }, spec)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fix test results for verifications

Revision ID: 37fdbb373e8d
Revises: 484cd9413e66
Create Date: 2016-12-29 19:54:23.804525

"""

from alembic import op
import sqlalchemy as sa

from rally.common.db import sa_types
from rally import exceptions

# revision identifiers, used by Alembic.
revision = "37fdbb373e8d"
down_revision = "484cd9413e66"
branch_labels = None
depends_on = None


verifications_helper = sa.Table(
    "verifications",
    sa.MetaData(),
    sa.Column("id", sa.Integer, primary_key=True, autoincrement=True),
    sa.Column("tests", sa_types.MutableJSONEncodedDict, default={})
)


def upgrade():
    connection = op.get_bind()
    for v in connection.execute(verifications_helper.select()):
        tests = v.tests
        for test in tests.values():
            duration = test.pop("time")
            test["duration"] = duration

        connection.execute(
            verifications_helper.update().where(
                verifications_helper.c.id == v.id).values(tests=tests))


def downgrade():
    raise exceptions.DowngradeNotSupported()



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
SLA (Service-level agreement) is set of details for determining compliance
with contracted values such as maximum error rate or minimum response time.
"""

from __future__ import division

from rally.common import streaming_algorithms
from rally import consts
from rally.task import sla
from rally.utils import strutils


@sla.configure(name="performance_degradation")
class PerformanceDegradation(sla.SLA):
    """Calculates performance degradation based on iteration time

    This SLA plugin finds minimum and maximum duration of
    iterations completed without errors during Rally task execution.
    Assuming that minimum duration is 100%, it calculates
    performance degradation against maximum duration.
    """
    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA7,
        "properties": {
            "max_degradation": {
                "type": "number",
                "minimum": 0.0,
            },
        },
        "required": [
            "max_degradation",
        ],
        "additionalProperties": False,
    }

    def __init__(self, criterion_value):
        super(PerformanceDegradation, self).__init__(criterion_value)
        self.max_degradation = self.criterion_value["max_degradation"]
        self.degradation = streaming_algorithms.DegradationComputation()

    def add_iteration(self, iteration):
        if not iteration.get("error"):
            self.degradation.add(iteration["duration"])
        self.success = self.degradation.result() <= self.max_degradation
        return self.success

    def merge(self, other):
        self.degradation.merge(other.degradation)
        self.success = self.degradation.result() <= self.max_degradation
        return self.success

    def details(self):
        res = strutils.format_float_to_str(self.degradation.result() or 0.0)
        return "Current degradation: %s%% - %s" % (res, self.status())



# Copyright 2016: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import io
import tempfile

import mock

from rally.cli import cliutils
from rally.cli.commands import verify
from rally.cli import envutils
from rally import consts
from rally import exceptions
from rally import plugins
from rally.verification import reporter
from tests.unit import fakes
from tests.unit import test


class VerifyCommandsTestCase(test.TestCase):

    def setUp(self):
        super(VerifyCommandsTestCase, self).setUp()

        self.verify = verify.VerifyCommands()
        self.fake_api = fakes.FakeAPI()

        self.deployment_name = "Some Deploy"
        self.deployment_uuid = "some-deploy-uuid"
        self.verifier_name = "My Verifier"
        self.verifier_uuid = "my-verifier-uuid"
        self.verifier_type = "OldSchoolTestTool"
        self.verifier_platform = "OpenStack"
        self.verification_uuid = "uuuiiiiddd"

        self.verifier_data = {
            "uuid": self.verifier_uuid,
            "name": self.verifier_name,
            "type": self.verifier_type,
            "platform": self.verifier_platform,
            "description": "The best tool in the world",
            "created_at": "2016-01-01T17:00:03",
            "updated_at": "2016-01-01T17:01:05",
            "status": "installed",
            "source": "https://example.com",
            "version": "master",
            "system_wide": False,
            "extra_settings": {},
            "manager.repo_dir": "./verifiers/repo",
            "manager.venv_dir": "./verifiers/.venv"
        }

        self.verification_data = {
            "uuid": self.verification_uuid,
            "verifier_uuid": self.verifier_uuid,
            "deployment_uuid": self.deployment_uuid,
            "tags": ["bar", "foo"],
            "status": "success",
            "created_at": "2016-01-01T17:00:03",
            "updated_at": "2016-01-01T17:01:05",
            "tests_count": 2,
            "tests_duration": 4,
            "success": 1,
            "skipped": 0,
            "expected_failures": 0,
            "unexpected_success": 0,
            "failures": 1,
            "run_args": {
                "load_list": ["test_1", "test_2"],
                "skip_list": ["test_3"],
                "concurrency": "3"},
            "tests": {
                "test_1": {
                    "name": "test_1",
                    "status": "success",
                    "duration": 2,
                    "tags": []},
                "test_2": {
                    "name": "test_2",
                    "status": "fail",
                    "duration": 2,
                    "traceback": "Some traceback"}
            },
            "test_2": {
                "name": "test_2",
                "status": "fail",
                "duration": 2,
                "traceback": "Some traceback"}
        }

        self.results_data = {
            "totals": {"tests_count": 2,
                       "tests_duration": 4,
                       "success": 1,
                       "skipped": 0,
                       "expected_failures": 0,
                       "unexpected_success": 0,
                       "failures": 1},
            "tests": {
                "test_1": {
                    "name": "test_1",
                    "status": "success",
                    "duration": 2,
                    "tags": []}
            },
            "test_2": {
                "name": "test_2",
                "status": "fail",
                "duration": 4,
                "tags": []}
        }

    @mock.patch("rally.cli.commands.verify.cliutils.print_list")
    @mock.patch("rally.cli.commands.verify.logging.is_debug",
                return_value=True)
    def test_list_plugins(self, mock_is_debug, mock_print_list):
        self.verify.list_plugins(self.fake_api, platform="some")
        self.fake_api.verifier.list_plugins.assert_called_once_with(
            platform="some")

    @mock.patch("rally.cli.commands.verify.fileutils.update_globals_file")
    def test_create_verifier(self, mock_update_globals_file):
        self.fake_api.verifier.create.return_value = self.verifier_uuid
        self.fake_api.verifier.get.return_value = self.verifier_data

        self.verify.create_verifier(self.fake_api, "a", vtype="b",
                                    platform="c", source="d", version="e",
                                    system_wide=True, extra={})
        self.fake_api.verifier.create.assert_called_once_with(
            name="a", vtype="b", platform="c", source="d", version="e",
            system_wide=True, extra_settings={})

        self.fake_api.verifier.get.assert_called_once_with(
            verifier_id=self.verifier_uuid)
        mock_update_globals_file.assert_called_once_with(
            envutils.ENV_VERIFIER, self.verifier_uuid)

    @mock.patch("rally.cli.commands.verify.fileutils.update_globals_file")
    def test_use_verifier(self, mock_update_globals_file):
        self.fake_api.verifier.get.return_value = self.verifier_data
        self.verify.use_verifier(self.fake_api, self.verifier_uuid)
        self.fake_api.verifier.get.assert_called_once_with(
            verifier_id=self.verifier_uuid)
        mock_update_globals_file.assert_called_once_with(
            envutils.ENV_VERIFIER, self.verifier_uuid)

    @mock.patch("rally.cli.commands.verify.cliutils.print_list")
    def test_list_verifiers_empty_verifiers(self, mock_print_list):
        self.fake_api.verifier.list.return_value = []
        self.verify.list_verifiers(self.fake_api)

        self.verify.list_verifiers(self.fake_api, "foo")
        self.verify.list_verifiers(self.fake_api)

        self.fake_api.verifier.list.assert_has_calls(
            [mock.call(status=None), mock.call(status="foo")])

    @mock.patch("rally.cli.commands.verify.cliutils.print_list")
    def test_list_verifiers(self, mock_print_list):
        self.fake_api.verifier.list.return_value = [self.verifier_data]

        additional_fields = ["UUID", "Name", "Type", "Platform", "Created at",
                             "Updated at", "Status", "Version", "System-wide",
                             "Active"]
        additional_keys = ["normalize_field_names", "sortby_index",
                           "formatters"]
        self.verify.list_verifiers(self.fake_api)
        # astarove: should be replaced on mock_print_list.assert_called_once()
        self.assertEqual(1, mock_print_list.call_count)
        self.assertEqual(([self.verifier_data], additional_fields),
                         mock_print_list.call_args[0])
        self.assertEqual(additional_keys.sort(),
                         list(mock_print_list.call_args[1].keys()).sort())

    @mock.patch("rally.cli.commands.verify.envutils.get_global")
    def test_show_verifier(self, mock_get_global):
        self.fake_api.verifier.get.return_value = self.verifier_data
        self.verify._base_dir = mock.Mock(return_value="./verifiers/")

        # It is a hard task to mock default value of function argument, so we
        # need to apply this workaround
        original_print_dict = cliutils.print_dict
        print_dict_calls = []

        def print_dict(*args, **kwargs):
            print_dict_calls.append(io.StringIO())
            kwargs["out"] = print_dict_calls[-1]
            original_print_dict(*args, **kwargs)

        with mock.patch.object(verify.cliutils, "print_dict",
                               new=print_dict):
            self.verify.show_verifier(self.fake_api, self.verifier_uuid)

        self.assertEqual(1, len(print_dict_calls))

        self.assertEqual(
            "+---------------------------------------------+\n"
            "|                  Verifier                   |\n"
            "+----------------+----------------------------+\n"
            "| UUID           | my-verifier-uuid           |\n"
            "| Status         | installed                  |\n"
            "| Created at     | 2016-01-01 17:00:03        |\n"
            "| Updated at     | 2016-01-01 17:01:05        |\n"
            "| Active         | -                          |\n"
            "| Name           | My Verifier                |\n"
            "| Description    | The best tool in the world |\n"
            "| Type           | OldSchoolTestTool          |\n"
            "| Platform       | OpenStack                  |\n"
            "| Source         | https://example.com        |\n"
            "| Version        | master                     |\n"
            "| System-wide    | False                      |\n"
            "| Extra settings | -                          |\n"
            "| Location       | ./verifiers/repo           |\n"
            "| Venv location  | ./verifiers/.venv          |\n"
            "+----------------+----------------------------+\n",
            print_dict_calls[0].getvalue())

        self.fake_api.verifier.get.assert_called_once_with(
            verifier_id=self.verifier_uuid)

    def test_delete_verifier(self):
        self.verify.delete_verifier(self.fake_api, "v_id", "d_id", force=True)
        self.fake_api.verifier.delete.assert_called_once_with(
            verifier_id="v_id", deployment_id="d_id", force=True)

    def test_update_verifier(self):
        self.verify.update_verifier(self.fake_api, self.verifier_uuid)
        self.assertFalse(self.fake_api.verifier.update.called)

        self.verify.update_verifier(self.fake_api, self.verification_uuid,
                                    update_venv=True,
                                    system_wide=True)
        self.assertFalse(self.fake_api.verifier.update.called)

        self.verify.update_verifier(self.fake_api, self.verification_uuid,
                                    system_wide=True,
                                    no_system_wide=True)
        self.assertFalse(self.fake_api.verifier.update.called)

        self.verify.update_verifier(self.fake_api, self.verification_uuid,
                                    version="a",
                                    system_wide=True)
        self.fake_api.verifier.update.assert_called_once_with(
            verifier_id=self.verification_uuid, system_wide=True,
            version="a", update_venv=None)

    @mock.patch("rally.cli.commands.verify.open", create=True)
    @mock.patch("rally.cli.commands.verify.os.path.exists")
    def test_configure_verifier(self, mock_exists, mock_open):
        self.verify.configure_verifier(self.fake_api, self.verifier_uuid,
                                       self.deployment_uuid,
                                       new_configuration="/p/a/t/h",
                                       reconfigure=True,
                                       show=True)
        self.assertFalse(self.fake_api.verifier.configure.called)

        mock_exists.return_value = False
        self.verify.configure_verifier(self.fake_api, self.verifier_uuid,
                                       self.deployment_uuid,
                                       new_configuration="/p/a/t/h",
                                       show=True)
        self.assertFalse(self.fake_api.verifier.override_configuration.called)

        mock_exists.return_value = True
        mock_open.return_value = mock.mock_open(read_data="data").return_value
        self.verify.configure_verifier(self.fake_api, self.verifier_uuid,
                                       self.deployment_uuid,
                                       new_configuration="/p/a/t/h",
                                       show=True)
        mock_open.assert_called_once_with("/p/a/t/h")
        self.fake_api.verifier.override_configuration(self.verifier_uuid,
                                                      self.deployment_uuid,
                                                      "data")

        tf = tempfile.NamedTemporaryFile()
        with open(tf.name, "w") as f:
            f.write("[DEFAULT]\nopt = val\n[foo]\nopt = val")
        self.verify.configure_verifier(self.fake_api, self.verifier_uuid,
                                       self.deployment_uuid,
                                       extra_options=tf.name)
        expected_options = {"foo": {"opt": "val"},
                            "DEFAULT": {"opt": "val"}}
        self.fake_api.verifier.configure.assert_called_once_with(
            verifier=self.verifier_uuid, deployment_id=self.deployment_uuid,
            extra_options=expected_options, reconfigure=False)

        self.verify.configure_verifier(self.fake_api, self.verifier_uuid,
                                       self.deployment_uuid,
                                       extra_options="{foo: {opt: val}, "
                                                     "DEFAULT: {opt: val}}")
        self.fake_api.verifier.configure.assert_called_with(
            verifier=self.verifier_uuid, deployment_id=self.deployment_uuid,
            extra_options=expected_options, reconfigure=False)

    def test_list_verifier_tests(self):
        self.fake_api.verifier.list_tests.return_value = ["test_1", "test_2"]
        self.verify.list_verifier_tests(self.fake_api, self.verifier_uuid, "p")

        self.fake_api.verifier.list_tests.return_value = []
        self.verify.list_verifier_tests(self.fake_api, self.verifier_uuid, "p")

        self.fake_api.verifier.list_tests.assert_has_calls(
            [mock.call(verifier_id=self.verifier_uuid, pattern="p"),
             mock.call(verifier_id=self.verifier_uuid, pattern="p")])

    def test_add_verifier_ext(self):
        self.verify.add_verifier_ext(self.fake_api, self.verifier_uuid,
                                     "a", "b", "c")
        self.fake_api.verifier.add_extension.assert_called_once_with(
            verifier_id=self.verifier_uuid,
            source="a", version="b", extra_settings="c")

    @mock.patch("rally.cli.commands.verify.cliutils.print_list")
    @mock.patch("rally.cli.commands.verify.logging.is_debug",
                return_value=True)
    def test_list_verifier_exts_empty_list(self,
                                           mock_is_debug, mock_print_list):
        self.fake_api.verifier.list_extensions.return_value = []
        self.verify.list_verifier_exts(self.fake_api, self.verifier_uuid)

        self.verify.list_verifier_exts(self.fake_api, self.verifier_uuid)

        self.fake_api.verifier.list_extensions.assert_has_calls(
            [mock.call(verifier_id=self.verifier_uuid),
             mock.call(verifier_id=self.verifier_uuid)])

    @mock.patch("rally.cli.commands.verify.cliutils.print_list")
    @mock.patch("rally.cli.commands.verify.logging.is_debug",
                return_value=False)
    def test_list_verifier_exts(self, mock_is_debug, mock_print_list):
        ver_exts = self.fake_api.verifier.list_extensions
        ver_exts.return_value = [mock.MagicMock()]
        fields = ["Name", "Entry point"]
        self.verify.list_verifier_exts(self.fake_api, self.verifier_uuid)

        self.verify.list_verifier_exts(self.fake_api, self.verifier_uuid)

        self.fake_api.verifier.list_extensions.assert_has_calls(
            [mock.call(verifier_id=self.verifier_uuid),
             mock.call(verifier_id=self.verifier_uuid)])

        mock_print_list.assert_called_with(ver_exts.return_value,
                                           fields,
                                           normalize_field_names=True)

    @mock.patch("rally.cli.commands.verify.cliutils.print_list")
    @mock.patch("rally.cli.commands.verify.logging.is_debug",
                return_value=True)
    def test_list_verifier_exts_with_logging(self,
                                             mock_is_debug, mock_print_list):
        ver_exts = self.fake_api.verifier.list_extensions
        ver_exts.return_value = [mock.MagicMock()]
        fields = ["Name", "Entry point", "Location"]
        self.verify.list_verifier_exts(self.fake_api, self.verifier_uuid)

        self.verify.list_verifier_exts(self.fake_api, self.verifier_uuid)

        self.fake_api.verifier.list_extensions.assert_has_calls(
            [mock.call(verifier_id=self.verifier_uuid),
             mock.call(verifier_id=self.verifier_uuid)])

        mock_print_list.assert_called_with(ver_exts.return_value,
                                           fields,
                                           normalize_field_names=True)

    def test_delete_verifier_ext(self):
        self.verify.delete_verifier_ext(self.fake_api, self.verifier_uuid,
                                        "ext_name")
        self.fake_api.verifier.delete_extension.assert_called_once_with(
            verifier_id=self.verifier_uuid, name="ext_name")

    @mock.patch("rally.cli.commands.verify.fileutils.update_globals_file")
    @mock.patch("rally.cli.commands.verify.os.path.exists")
    def test_start(self, mock_exists, mock_update_globals_file):
        self.verify.start(self.fake_api, self.verifier_uuid,
                          self.deployment_uuid, pattern="pattern",
                          load_list="load-list")
        self.assertFalse(self.fake_api.verification.start.called)

        verification = self.verification_data
        self.fake_api.verification.start.return_value = {
            "verification": verification,
            "totals": self.results_data["totals"],
            "tests": self.results_data["tests"]}
        self.fake_api.verification.get.return_value = verification

        mock_exists.return_value = False
        self.verify.start(self.fake_api, self.verifier_uuid,
                          self.deployment_uuid, load_list="/p/a/t/h")
        self.assertFalse(self.fake_api.verification.start.called)

        mock_exists.return_value = True
        tf = tempfile.NamedTemporaryFile()
        with open(tf.name, "w") as f:
            f.write("test_1\ntest_2")
        self.verify.start(self.fake_api, self.verifier_uuid,
                          self.deployment_uuid, tags=["foo"],
                          load_list=tf.name)
        self.fake_api.verification.start.assert_called_once_with(
            verifier_id=self.verifier_uuid,
            deployment_id=self.deployment_uuid,
            tags=["foo"], load_list=["test_1", "test_2"])

        mock_exists.return_value = False
        self.fake_api.verification.start.reset_mock()

        self.verify.start(self.fake_api, self.verifier_uuid,
                          self.verifier_uuid, skip_list="/p/a/t/h")
        self.assertFalse(self.fake_api.verification.start.called)

        tf = tempfile.NamedTemporaryFile()
        with open(tf.name, "w") as f:
            f.write("test_1:\ntest_2: Reason\n")
        mock_exists.return_value = True
        self.verify.start(self.fake_api, self.verifier_uuid,
                          self.deployment_uuid, skip_list=tf.name)
        self.fake_api.verification.start.assert_called_once_with(
            verifier_id=self.verifier_uuid,
            deployment_id=self.deployment_uuid,
            tags=None,
            skip_list={"test_1": None, "test_2": "Reason"})

        mock_exists.return_value = False
        self.fake_api.verification.start.reset_mock()
        self.verify.start(self.fake_api, self.verifier_uuid,
                          self.deployment_uuid, xfail_list="/p/a/t/h")
        self.assertFalse(self.fake_api.verification.start.called)

        tf = tempfile.NamedTemporaryFile()
        with open(tf.name, "w") as f:
            f.write("test_1:\ntest_2: Reason\n")
        mock_exists.return_value = True
        self.verify.start(self.fake_api, self.verifier_uuid,
                          self.deployment_uuid, xfail_list=tf.name)
        self.fake_api.verification.start.assert_called_once_with(
            verifier_id=self.verifier_uuid,
            deployment_id=self.deployment_uuid, tags=None,
            xfail_list={"test_1": None, "test_2": "Reason"})

        self.fake_api.verification.get.assert_called_with(
            verification_uuid=self.verification_uuid)
        mock_update_globals_file.assert_called_with(
            envutils.ENV_VERIFICATION, self.verification_uuid)

        self.fake_api.verification.get.reset_mock()
        mock_update_globals_file.reset_mock()
        self.verify.start(self.fake_api, self.verifier_uuid,
                          self.deployment_uuid, detailed=True,
                          do_use=False)
        self.assertFalse(self.fake_api.verification.get.called)
        self.assertFalse(mock_update_globals_file.called)

    @mock.patch("rally.cli.commands.verify.os.path.exists")
    @mock.patch("rally.cli.commands.verify.fileutils.update_globals_file")
    def test_start_on_unfinished_deployment(self, mock_update_globals_file,
                                            mock_exists):
        deployment_id = self.deployment_uuid
        deployment_name = self.deployment_name
        exc = exceptions.DeploymentNotFinishedStatus(
            name=deployment_name,
            uuid=deployment_id,
            status=consts.DeployStatus.DEPLOY_INIT)
        self.fake_api.verification.start.side_effect = exc
        self.assertEqual(
            1, self.verify.start(self.fake_api,
                                 self.deployment_uuid, deployment_id))

    @mock.patch("rally.cli.commands.verify.fileutils.update_globals_file")
    def test_use(self, mock_update_globals_file):
        self.fake_api.verification.get.return_value = self.verification_data
        self.verify.use(self.fake_api, self.verification_uuid)
        self.fake_api.verification.get.assert_called_once_with(
            verification_uuid=self.verification_uuid)
        mock_update_globals_file.assert_called_once_with(
            envutils.ENV_VERIFICATION, self.verification_uuid)

    @mock.patch("rally.cli.commands.verify.fileutils.update_globals_file")
    def test_rerun(self, mock_update_globals_file):
        self.fake_api.verification.rerun.return_value = {
            "verification": self.verification_data,
            "totals": self.results_data["totals"],
            "tests": self.results_data["tests"]}
        self.fake_api.verification.get.return_value = self.verification_data

        self.verify.rerun(self.fake_api, self.verification_uuid,
                          self.deployment_uuid, failed=True)
        self.fake_api.verification.rerun.assert_called_once_with(
            verification_uuid=self.verification_uuid,
            concurrency=None,
            deployment_id="some-deploy-uuid",
            failed=True, tags=None)
        mock_update_globals_file.assert_called_once_with(
            envutils.ENV_VERIFICATION, self.verification_uuid)

    def test_show(self):

        verification = self.verification_data
        self.fake_api.verifier.get.return_value = self.verifier_data
        self.fake_api.verification.get.return_value = verification
        self.fake_api.deployment.get.return_value = {
            "name": self.deployment_name, "uuid": self.deployment_uuid}

        # It is a hard task to mock default value of function argument, so we
        # need to apply this workaround
        original_print_dict = cliutils.print_dict
        print_dict_calls = []

        def print_dict(*args, **kwargs):
            print_dict_calls.append(io.StringIO())
            kwargs["out"] = print_dict_calls[-1]
            original_print_dict(*args, **kwargs)

        with mock.patch.object(verify.cliutils, "print_dict",
                               new=print_dict):
            self.verify.show(self.fake_api, self.verifier_uuid, detailed=True)

        self.assertEqual(1, len(print_dict_calls))

        self.assertEqual(
            "+----------------------------------------------------------------"
            "--------------------+\n"
            "|                                    Verification                "
            "                    |\n"
            "+---------------------+------------------------------------------"
            "--------------------+\n"
            "| UUID                | uuuiiiiddd                               "
            "                    |\n"
            "| Status              | success                                  "
            "                    |\n"
            "| Started at          | 2016-01-01 17:00:03                      "
            "                    |\n"
            "| Finished at         | 2016-01-01 17:01:05                      "
            "                    |\n"
            "| Duration            | 0:01:02                                  "
            "                    |\n"
            "| Run arguments       | concurrency: 3                           "
            "                    |\n"
            "|                     | load_list: (value is too long, will be di"
            "splayed separately) |\n"
            "|                     | skip_list: (value is too long, will be di"
            "splayed separately) |\n"
            "| Tags                | bar, foo                                 "
            "                    |\n"
            "| Verifier name       | My Verifier (UUID: my-verifier-uuid)     "
            "                    |\n"
            "| Verifier type       | OldSchoolTestTool (platform: OpenStack)  "
            "                    |\n"
            "| Deployment name     | Some Deploy (UUID: some-deploy-uuid)     "
            "                    |\n"
            "| Tests count         | 2                                        "
            "                    |\n"
            "| Tests duration, sec | 4                                        "
            "                    |\n"
            "| Success             | 1                                        "
            "                    |\n"
            "| Skipped             | 0                                        "
            "                    |\n"
            "| Expected failures   | 0                                        "
            "                    |\n"
            "| Unexpected success  | 0                                        "
            "                    |\n"
            "| Failures            | 1                                        "
            "                    |\n"
            "+---------------------+------------------------------------------"
            "--------------------+\n", print_dict_calls[0].getvalue())

        self.fake_api.verification.get.assert_called_once_with(
            verification_uuid=self.verifier_uuid)

        with mock.patch.object(verify.cliutils, "print_dict",
                               new=print_dict):
            self.verify.show(self.fake_api, self.verifier_uuid, detailed=False)
        self.assertEqual(2, len(print_dict_calls))

        self.assertEqual(
            print_dict_calls[1].getvalue(),
            "+---------------------------------------------------"
            "--------------------------------------+\n"
            "|                                      Verification "
            "                                      |\n"
            "+---------------------+-----------------------------"
            "--------------------------------------+\n"
            "| UUID                | uuuiiiiddd                  "
            "                                      |\n"
            "| Status              | success                     "
            "                                      |\n"
            "| Started at          | 2016-01-01 17:00:03         "
            "                                      |\n"
            "| Finished at         | 2016-01-01 17:01:05         "
            "                                      |\n"
            "| Duration            | 0:01:02                     "
            "                                      |\n"
            "| Run arguments       | concurrency: 3              "
            "                                      |\n"
            "|                     | load_list: (value is too lon"
            "g, use 'detailed' flag to display it) |\n"
            "|                     | skip_list: (value is too lon"
            "g, use 'detailed' flag to display it) |\n"
            "| Tags                | bar, foo                    "
            "                                      |\n"
            "| Verifier name       | My Verifier (UUID: my-verifi"
            "er-uuid)                              |\n"
            "| Verifier type       | OldSchoolTestTool (platform:"
            " OpenStack)                           |\n"
            "| Deployment name     | Some Deploy (UUID: some-depl"
            "oy-uuid)                              |\n"
            "| Tests count         | 2                           "
            "                                      |\n"
            "| Tests duration, sec | 4                           "
            "                                      |\n"
            "| Success             | 1                           "
            "                                      |\n"
            "| Skipped             | 0                           "
            "                                      |\n"
            "| Expected failures   | 0                           "
            "                                      |\n"
            "| Unexpected success  | 0                           "
            "                                      |\n"
            "| Failures            | 1                           "
            "                                      |\n"
            "+---------------------+-----------------------------"
            "--------------------------------------+\n",
        )

        self.fake_api.verification.get.assert_called_with(
            verification_uuid=self.verifier_uuid)

    @mock.patch("rally.cli.commands.verify.cliutils.print_list")
    def test_list_empty_verifications(self, mock_print_list):
        self.fake_api.verification.list.return_value = []
        self.verify.list(self.fake_api, self.verifier_uuid,
                         self.deployment_uuid)

        self.verify.list(self.fake_api, self.verifier_uuid,
                         self.deployment_uuid, "foo", "bar")
        self.verify.list(self.fake_api)

        self.fake_api.verification.list.assert_has_calls(
            [mock.call(verifier_id=self.verifier_uuid,
                       deployment_id=self.deployment_uuid,
                       tags=None, status=None),
             mock.call(verifier_id=self.verifier_uuid,
                       deployment_id=self.deployment_uuid,
                       tags="foo", status="bar"),
             mock.call(verifier_id=None, deployment_id=None,
                       tags=None, status=None)])

    @mock.patch("rally.cli.commands.verify.cliutils.print_list")
    def test_list(self, mock_print_list):
        self.fake_api.verification.list.return_value = [self.verification_data]
        self.verify.list(self.fake_api, self.verifier_uuid,
                         self.deployment_uuid)

        additional_fields = ["UUID", "Tags", "Verifier name",
                             "Deployment name", "Started at", "Finished at",
                             "Duration", "Status"]
        additional_keys = ["normalize_field_names", "sortby_index",
                           "formatters"]
        # astarove: Should be replaced on mock_print_list.assert_called_once())
        self.assertEqual(1, mock_print_list.call_count)
        self.assertEqual(([self.verification_data], additional_fields),
                         mock_print_list.call_args[0])
        self.assertEqual(additional_keys.sort(),
                         list(mock_print_list.call_args[1].keys()).sort())

    def test_delete(self):
        self.verify.delete(self.fake_api, "v_uuid")
        self.fake_api.verification.delete.assert_called_once_with(
            verification_uuid="v_uuid")

        self.verify.delete(self.fake_api, ["v1_uuid", "v2_uuid"])
        self.fake_api.verification.delete.assert_has_calls(
            [mock.call(verification_uuid="v1_uuid"),
             mock.call(verification_uuid="v2_uuid")])

    @mock.patch("rally.cli.commands.verify.os")
    @mock.patch("rally.cli.commands.verify.webbrowser.open_new_tab")
    @mock.patch("rally.cli.commands.verify.open", create=True)
    def test_report(self, mock_open, mock_open_new_tab, mock_os):
        output_dest = "/p/a/t/h"
        output_type = "type"
        content = "content"
        self.fake_api.verification.report.return_value = {
            "files": {output_dest: content}, "open": output_dest}
        mock_os.path.exists.return_value = False

        self.verify.report(self.fake_api,
                           verification_uuid=self.verifier_uuid,
                           output_type=output_type,
                           output_dest=output_dest, open_it=True)
        self.fake_api.verification.report.assert_called_once_with(
            uuids=[self.verifier_uuid], output_type=output_type,
            output_dest=output_dest)
        mock_open.assert_called_once_with(mock_os.path.abspath.return_value,
                                          "w")
        mock_os.makedirs.assert_called_once_with(
            mock_os.path.dirname.return_value)

        mock_open.reset_mock()
        mock_open_new_tab.reset_mock()
        mock_os.makedirs.reset_mock()

        mock_os.path.exists.return_value = True
        self.fake_api.verification.report.return_value = {
            "files": {output_dest: content}, "print": "foo"}

        self.verify.report(self.fake_api, self.verifier_uuid,
                           output_type=output_type,
                           output_dest=output_dest)

        self.assertFalse(mock_open_new_tab.called)
        self.assertFalse(mock_os.makedirs.called)

    @mock.patch("rally.cli.commands.verify.VerifyCommands.use")
    @mock.patch("rally.cli.commands.verify.open", create=True)
    @mock.patch("rally.cli.commands.verify.os.path.exists")
    def test_import_results(self, mock_exists, mock_open, mock_use):
        mock_exists.return_value = False
        self.verify.import_results(self.fake_api, self.verifier_uuid,
                                   self.deployment_uuid)
        self.assertFalse(self.fake_api.verification.import_results.called)

        verification = self.verification_data
        results = self.results_data
        self.fake_api.verification.import_results.return_value = (
            verification, results)

        mock_exists.return_value = True
        mock_open.return_value = mock.mock_open(read_data="data").return_value
        self.verify.import_results(self.fake_api,
                                   verifier_id=self.verifier_uuid,
                                   deployment=self.deployment_uuid,
                                   file_to_parse="/p/a/t/h")
        mock_open.assert_called_once_with("/p/a/t/h", "r")
        self.fake_api.verification.import_results.assert_called_once_with(
            verifier_id=self.verifier_uuid,
            deployment_id=self.deployment_uuid,
            data="data")

        mock_use.assert_called_with(self.fake_api, self.verification_uuid)

        mock_use.reset_mock()
        self.verify.import_results(self.fake_api, "v_id", "d_id", do_use=False)
        self.assertFalse(mock_use.called)

    @plugins.ensure_plugins_are_loaded
    def test_default_reporters(self):
        available_reporters = {
            cls.get_name().lower()
            for cls in reporter.VerificationReporter.get_all()
            # ignore possible external plugins
            if cls.__module__.startswith("rally")}
        listed_in_cli = {name.lower() for name in verify.DEFAULT_REPORT_TYPES}
        not_listed = available_reporters - listed_in_cli

        if not_listed:
            self.fail("All default reporters should be listed in "
                      "%s.DEFAULTS_REPORTERS (case of letters doesn't matter)."
                      " Missed reporters: %s" % (verify.__name__,
                                                 ", ".join(not_listed)))



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add config field to env models

Revision ID: 95208e4eface
Revises: 7287df262dbc
Create Date: 2018-02-04 13:48:35.779255

"""

from alembic import op
import sqlalchemy as sa

from rally.common.db import sa_types
from rally import exceptions


# revision identifiers, used by Alembic.
revision = "95208e4eface"
down_revision = "7287df262dbc"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("envs") as batch_op:
        batch_op.add_column(
            sa.Column("config", sa_types.MutableJSONEncodedDict))


def downgrade():
    raise exceptions.DowngradeNotSupported()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import ddt
import mock

from rally.plugins.common.scenarios.dummy import dummy
from tests.unit import test


DUMMY = "rally.plugins.common.scenarios.dummy.dummy."


@ddt.ddt
class DummyFailureTestCase(test.TestCase):

    @ddt.data({"iteration": 0, "kwargs": {}},
              {"iteration": 0, "kwargs": {"each": 1}},
              {"iteration": 5, "kwargs": {"from_iteration": 4},
               "raises": False},
              {"iteration": 5,
               "kwargs": {"from_iteration": 5, "to_iteration": 5}},
              {"iteration": 5,
               "kwargs": {"from_iteration": 4, "to_iteration": 5}},
              {"iteration": 5,
               "kwargs": {"from_iteration": 5, "to_iteration": 6}},
              {"iteration": 5,
               "kwargs": {"from_iteration": 4, "to_iteration": 6}},
              {"iteration": 5, "kwargs": {"from_iteration": 4,
                                          "to_iteration": 6, "sleep": 5}},
              {"iteration": 5, "raises": False,
               "kwargs": {"from_iteration": 4, "to_iteration": 6,
                          "sleep": 5, "each": 2}},
              {"iteration": 6, "kwargs": {"from_iteration": 4,
                                          "to_iteration": 6,
                                          "sleep": 5, "each": 2}})
    @ddt.unpack
    @mock.patch(DUMMY + "utils.interruptable_sleep")
    def test_run(self, mock_interruptable_sleep, iteration, kwargs,
                 raises=True):
        scenario = dummy.DummyFailure(
            test.get_test_context(iteration=iteration))
        if raises:
            self.assertRaises(dummy.DummyScenarioException, scenario.run,
                              **kwargs)
        else:
            scenario.run(**kwargs)
        mock_interruptable_sleep.assert_called_once_with(
            kwargs.get("sleep", 0.1))


@ddt.ddt
class DummyTestCase(test.TestCase):

    @mock.patch(DUMMY + "utils.interruptable_sleep")
    def test_dummy(self, mock_interruptable_sleep):
        scenario = dummy.Dummy(test.get_test_context())
        scenario.sleep_between = mock.MagicMock()

        scenario.run(sleep=10)
        mock_interruptable_sleep.assert_called_once_with(10)

    @mock.patch(DUMMY + "utils.interruptable_sleep")
    def test_dummy_exception(self, mock_interruptable_sleep):
        scenario = dummy.DummyException(test.get_test_context())

        size_of_message = 5
        self.assertRaises(dummy.DummyScenarioException,
                          scenario.run, size_of_message, sleep=10)
        mock_interruptable_sleep.assert_called_once_with(10)

    def test_dummy_exception_probability(self):
        scenario = dummy.DummyExceptionProbability(test.get_test_context())

        # should not raise an exception as probability is 0
        for i in range(100):
            scenario.run(exception_probability=0)

        # should always raise an exception as probability is 1
        for i in range(100):
            self.assertRaises(dummy.DummyScenarioException,
                              scenario.run,
                              exception_probability=1)

    @mock.patch(DUMMY + "random")
    def test_dummy_output(self, mock_random):
        mock_random.randint.side_effect = lambda min_, max_: max_
        desc = "This is a description text for %s"
        for random_range, exp in (None, 25), (1, 1), (42, 42):
            scenario = dummy.DummyOutput(test.get_test_context())
            if random_range is None:
                scenario.run()
            else:
                scenario.run(random_range=random_range)
            expected = {
                "additive": [
                    {"chart_plugin": "StatsTable",
                     "data": [["%s stat" % s, exp]
                              for s in ("foo", "bar", "spam")],
                     "description": desc % "Additive StatsTable",
                     "title": "Additive StatsTable"},
                    {"chart_plugin": "StackedArea",
                     "data": [["foo %i" % i, exp] for i in range(1, 7)],
                     "label": "Measure this in Foo units",
                     "title": "Additive StackedArea (no description)"},
                    {"chart_plugin": "Lines",
                     "data": [["bar %i" % i, exp] for i in range(1, 4)],
                     "description": desc % "Additive Lines",
                     "label": "Measure this in Bar units",
                     "title": "Additive Lines"},
                    {"chart_plugin": "Pie",
                     "data": [["spam %i" % i, exp] for i in range(1, 4)],
                     "description": desc % "Additive Pie",
                     "title": "Additive Pie"}],
                "complete": [
                    {"axis_label": "This is a custom X-axis label",
                     "chart_plugin": "Lines",
                     "data": [["Foo", [[i, exp] for i in range(1, 8)]],
                              ["Bar", [[i, exp] for i in range(1, 8)]],
                              ["Spam", [[i, exp] for i in range(1, 8)]]],
                     "description": desc % "Complete Lines",
                     "label": "Measure this is some units",
                     "title": "Complete Lines"},
                    {"axis_label": "This is a custom X-axis label",
                     "chart_plugin": "StackedArea",
                     "data": [["alpha", [[i, exp] for i in range(50)]],
                              ["beta", [[i, exp] for i in range(50)]],
                              ["gamma", [[i, exp] for i in range(50)]]],
                     "description": desc % "Complete StackedArea",
                     "label": "Yet another measurement units",
                     "title": "Complete StackedArea"},
                    {"title": "Arbitrary Text",
                     "chart_plugin": "TextArea",
                     "data": ["Lorem ipsum dolor sit amet, consectetur "
                              "adipiscing elit, sed do eiusmod tempor "
                              "incididunt ut labore et dolore magna "
                              "aliqua." * 2] * 4},
                    {"chart_plugin": "Pie",
                     "data": [[s, exp] for s in ("delta", "epsilon", "zeta",
                                                 "theta", "lambda", "omega")],
                     "title": "Complete Pie (no description)"},
                    {"chart_plugin": "Table",
                     "data": {"cols": ["%s column" % s
                                       for s in ("mu", "xi", "pi",
                                                 "tau", "chi")],
                              "rows": [["%s row" % s, exp, exp, exp, exp]
                                       for s in ("iota", "nu", "rho",
                                                 "phi", "psi")]},
                     "description": desc % "Complete Table",
                     "title": "Complete Table"}]}

        self.assertEqual(expected, scenario._output)

    def test_dummy_random_fail_in_atomic(self):
        scenario = dummy.DummyRandomFailInAtomic(test.get_test_context())

        for i in range(10):
            scenario.run(exception_probability=0)
        for i in range(10):
            self.assertRaises(KeyError,
                              scenario.run,
                              exception_probability=1)

    @ddt.data({},
              {"actions_num": 5, "sleep_min": 0, "sleep_max": 2},
              {"actions_num": 7, "sleep_min": 1.23, "sleep_max": 4.56},
              {"actions_num": 1, "sleep_max": 4.56},
              {"sleep_min": 1})
    @ddt.unpack
    @mock.patch(DUMMY + "random")
    @mock.patch(DUMMY + "utils.interruptable_sleep")
    def test_dummy_random_action(self, mock_interruptable_sleep, mock_random,
                                 **kwargs):
        mock_random.uniform.side_effect = range(100)

        scenario = dummy.DummyRandomAction(test.get_test_context())
        scenario.run(**kwargs)
        actions_num = kwargs.get("actions_num", 5)
        calls = [mock.call(i) for i in range(actions_num)]
        self.assertEqual(calls, mock_interruptable_sleep.mock_calls)

        calls = [mock.call(kwargs.get("sleep_min", 0),
                           kwargs.get("sleep_max", 0))
                 for i in range(actions_num)]
        self.assertEqual(calls, mock_random.uniform.mock_calls)
        for i in range(actions_num):
            self._test_atomic_action_timer(scenario.atomic_actions(),
                                           "action_%d" % i)

    @ddt.data({"number_of_actions": 5, "sleep_factor": 1},
              {"number_of_actions": 7, "sleep_factor": 2},
              {"number_of_actions": 1, "sleep_factor": 3})
    @ddt.unpack
    @mock.patch(DUMMY + "utils.interruptable_sleep")
    def test_dummy_timed_atomic_actions(self, mock_interruptable_sleep,
                                        number_of_actions, sleep_factor):
        dummy.DummyRandomAction(test.get_test_context()).run(
            number_of_actions, sleep_factor)
        scenario = dummy.DummyTimedAtomicAction(test.get_test_context())
        scenario.run(number_of_actions, sleep_factor)
        for i in range(number_of_actions):
            self._test_atomic_action_timer(scenario.atomic_actions(),
                                           "action_%d" % i)
            mock_interruptable_sleep.assert_any_call(i * sleep_factor)



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for db.task layer."""

import collections
import datetime as dt

import ddt
import mock

from rally.common import objects
from rally import consts
from rally import exceptions
from tests.unit import test


@ddt.ddt
class TaskTestCase(test.TestCase):
    def setUp(self):
        super(TaskTestCase, self).setUp()
        self.task = {
            "uuid": "00ef46a2-c5b8-4aea-a5ca-0f54a10cbca1",
            "status": consts.TaskStatus.INIT,
            "validation_result": {},
        }

    @mock.patch("rally.common.objects.task.db.task_create")
    def test_init_with_create(self, mock_task_create):
        mock_task_create.return_value = self.task
        task = objects.Task(status=consts.TaskStatus.CRASHED)
        mock_task_create.assert_called_once_with({
            "status": consts.TaskStatus.CRASHED})
        self.assertEqual(task["uuid"], self.task["uuid"])

    @mock.patch("rally.common.objects.task.db.task_create")
    def test_init_without_create(self, mock_task_create):
        task = objects.Task(task=self.task)
        self.assertFalse(mock_task_create.called)
        self.assertEqual(task["uuid"], self.task["uuid"])

    @mock.patch("rally.common.objects.task.uuid.uuid4",
                return_value="some_uuid")
    @mock.patch("rally.common.objects.task.db.task_create")
    def test_init_with_fake_true(self, mock_task_create, mock_uuid4):
        task = objects.Task(temporary=True)
        self.assertFalse(mock_task_create.called)
        self.assertTrue(mock_uuid4.called)
        self.assertEqual(task["uuid"], mock_uuid4.return_value)

    @mock.patch("rally.common.db.api.task_get")
    def test_get(self, mock_task_get):
        mock_task_get.return_value = self.task
        task = objects.Task.get(self.task["uuid"])
        mock_task_get.assert_called_once_with(self.task["uuid"],
                                              detailed=False)
        self.assertEqual(task["uuid"], self.task["uuid"])

    @mock.patch("rally.common.objects.task.db.task_get_status")
    def test_get_status(self, mock_task_get_status):
        task = objects.Task(task=self.task)
        status = task.get_status(task["uuid"])
        self.assertEqual(status, mock_task_get_status.return_value)

    @mock.patch("rally.common.objects.task.db.task_delete")
    @mock.patch("rally.common.objects.task.db.task_create")
    def test_create_and_delete(self, mock_task_create, mock_task_delete):
        mock_task_create.return_value = self.task
        task = objects.Task()
        task.delete()
        mock_task_delete.assert_called_once_with(
            self.task["uuid"], status=None)

    @mock.patch("rally.common.objects.task.db.task_delete")
    @mock.patch("rally.common.objects.task.db.task_create")
    def test_create_and_delete_status(self, mock_task_create,
                                      mock_task_delete):
        mock_task_create.return_value = self.task
        task = objects.Task()
        task.delete(status=consts.TaskStatus.FINISHED)
        mock_task_delete.assert_called_once_with(
            self.task["uuid"], status=consts.TaskStatus.FINISHED)

    @mock.patch("rally.common.objects.task.db.task_delete")
    def test_delete_by_uuid(self, mock_task_delete):
        objects.Task.delete_by_uuid(self.task["uuid"])
        mock_task_delete.assert_called_once_with(
            self.task["uuid"], status=None)

    @mock.patch("rally.common.objects.task.db.task_delete")
    def test_delete_by_uuid_status(self, mock_task_delete):
        objects.Task.delete_by_uuid(self.task["uuid"],
                                    consts.TaskStatus.FINISHED)
        mock_task_delete.assert_called_once_with(
            self.task["uuid"], status=consts.TaskStatus.FINISHED)

    @mock.patch("rally.common.objects.task.db.task_list",
                return_value=[{"uuid": "a",
                               "created_at": "b",
                               "status": consts.TaskStatus.CRASHED,
                               "tag": "d",
                               "deployment_name": "some_name"}])
    def list(self, mock_db_task_list):
        tasks = objects.Task.list(status="somestatus")
        mock_db_task_list.assert_called_once_with("somestatus", None)
        self.assertIs(type(tasks), list)
        self.assertIsInstance(tasks[0], objects.Task)
        self.assertEqual(mock_db_task_list.return_value["uuis"],
                         tasks[0]["uuid"])

    @mock.patch("rally.common.objects.task.db.task_update")
    @mock.patch("rally.common.objects.task.db.task_create")
    def test_update(self, mock_task_create, mock_task_update):
        mock_task_create.return_value = self.task
        mock_task_update.return_value = {"opt": "val2"}
        deploy = objects.Task(opt="val1")
        deploy._update({"opt": "val2"})
        mock_task_update.assert_called_once_with(
            self.task["uuid"], {"opt": "val2"})
        self.assertEqual(deploy["opt"], "val2")

    @ddt.data(
        {
            "status": "some_status", "allowed_statuses": ("s_1", "s_2")
        },
        {
            "status": "some_status", "allowed_statuses": None
        }
    )
    @ddt.unpack
    @mock.patch("rally.common.objects.task.db.task_update_status")
    @mock.patch("rally.common.objects.task.db.task_update")
    def test_update_status(self, mock_task_update, mock_task_update_status,
                           status, allowed_statuses):
        task = objects.Task(task=self.task)
        task.update_status(consts.TaskStatus.FINISHED, allowed_statuses)
        if allowed_statuses:
            self.assertFalse(mock_task_update.called)
            mock_task_update_status.assert_called_once_with(
                self.task["uuid"],
                consts.TaskStatus.FINISHED,
                allowed_statuses
            )
        else:
            self.assertFalse(mock_task_update_status.called)
            mock_task_update.assert_called_once_with(
                self.task["uuid"],
                {"status": consts.TaskStatus.FINISHED},
            )

    @mock.patch("rally.common.objects.task.db.task_update")
    def test_update_verification_log(self, mock_task_update):
        mock_task_update.return_value = self.task
        task = objects.Task(task=self.task)
        task.set_validation_failed({"a": "fake"})
        mock_task_update.assert_called_once_with(
            self.task["uuid"],
            {"status": consts.TaskStatus.VALIDATION_FAILED,
             "validation_result": {"a": "fake"}}
        )

    @mock.patch("rally.common.objects.task.db.env_get")
    def test_to_dict(self, mock_env_get):
        workloads = [{"created_at": dt.datetime.now(),
                      "updated_at": dt.datetime.now()}]
        self.task.update({"env_uuid": "deployment_uuid",
                          "deployment_uuid": "deployment_uuid",
                          "created_at": dt.datetime.now(),
                          "updated_at": dt.datetime.now()})

        mock_env_get.return_value = {"name": "deployment_name"}

        task = objects.Task(task=self.task)
        serialized_task = task.to_dict()

        mock_env_get.assert_called_once_with(self.task["env_uuid"])
        self.assertEqual(self.task, serialized_task)

        self.task["subtasks"] = [{"workloads": workloads}]

    @mock.patch("rally.common.db.api.task_get")
    def test_get_detailed(self, mock_task_get):
        mock_task_get.return_value = {"results": [{
            "created_at": dt.datetime.now(),
            "updated_at": dt.datetime.now()}]}
        task_detailed = objects.Task.get("task_id", detailed=True)
        mock_task_get.assert_called_once_with("task_id", detailed=True)
        self.assertEqual(mock_task_get.return_value, task_detailed.task)

    @mock.patch("rally.common.objects.task.db.task_update")
    def test_set_failed(self, mock_task_update):
        mock_task_update.return_value = self.task
        task = objects.Task(task=self.task)
        task.set_failed("foo_type", "foo_error_message", "foo_trace")
        mock_task_update.assert_called_once_with(
            self.task["uuid"],
            {"status": consts.TaskStatus.CRASHED,
             "validation_result": {"etype": "foo_type",
                                   "msg": "foo_error_message",
                                   "trace": "foo_trace"}},
        )

    @mock.patch("rally.common.objects.task.Subtask")
    def test_add_subtask(self, mock_subtask):
        task = objects.Task(task=self.task)
        subtask = task.add_subtask(title="foo")
        mock_subtask.assert_called_once_with(
            self.task["uuid"], title="foo", contexts=None, description=None)
        self.assertIs(subtask, mock_subtask.return_value)

    @ddt.data(
        {
            "soft": True, "status": consts.TaskStatus.INIT
        },
        {
            "soft": True, "status": consts.TaskStatus.ABORTED
        },
        {
            "soft": True, "status": consts.TaskStatus.FINISHED
        },
        {
            "soft": True, "status": consts.TaskStatus.CRASHED
        },
        {
            "soft": False, "status": consts.TaskStatus.ABORTED
        },
        {
            "soft": False, "status": consts.TaskStatus.FINISHED
        },
        {
            "soft": False, "status": consts.TaskStatus.CRASHED
        }
    )
    @ddt.unpack
    def test_abort_with_finished_states(self, soft, status):
        task = objects.Task(mock.MagicMock(), fake=True)
        task.get_status = mock.MagicMock(return_value=status)
        task.update_status = mock.MagicMock()

        self.assertRaises(exceptions.RallyException, task.abort, soft)

        self.assertEqual(1, task.get_status.call_count)
        self.assertFalse(task.update_status.called)

    @ddt.data(True, False)
    def test_abort_with_running_state(self, soft):
        task = objects.Task(mock.MagicMock(), fake=True)
        task.get_status = mock.MagicMock(return_value="running")
        task.update_status = mock.MagicMock()

        task.abort(soft)
        if soft:
            status = consts.TaskStatus.SOFT_ABORTING
        else:
            status = consts.TaskStatus.ABORTING

        task.update_status.assert_called_once_with(
            status,
            allowed_statuses=(consts.TaskStatus.RUNNING,
                              consts.TaskStatus.SOFT_ABORTING)
        )

    @ddt.data(
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "output": {"additive": [], "complete": []},
                  "error": ["err1", "err2"], "atomic_actions": []},
         "expected": True},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": [], "complete": []},
                  "atomic_actions": [{"name": "foo", "started_at": 1.0,
                                      "finished_at": 5.2, "children": []}]},
         "expected": True},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": ["a1", "a2"],
                                          "complete": ["c1", "c2"]},
                  "atomic_actions": [{"name": "foo", "started_at": 1.0,
                                      "finished_at": 5.2, "children": []}]},
         "validate_output_calls": [("additive", "a1"), ("additive", "a2"),
                                   ("complete", "c1"), ("complete", "c2")],
         "expected": True},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": ["a1", "a2"],
                                          "complete": ["c1", "c2"]},
                  "atomic_actions": [{"name": "foo", "started_at": 1.0,
                                      "finished_at": 5.2, "children": []}]},
         "validate_output_return_value": "validation error message"},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [42], "output": {"additive": [], "complete": []},
                  "atomic_actions": [{"name": "foo", "started_at": 1.0,
                                      "finished_at": 5.2, "children": []}]}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": [], "complete": []},
                  "atomic_actions": [{"name": "foo", "started_at": 10,
                                      "finished_at": 52, "children": []}]}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": [], "complete": []},
                  "atomic_actions": [{"name": "non-float", "started_at": 1.0,
                                      "children": []}]}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": [], "complete": []},
                  "atomic_actions": [{"name": "foo", "started_at": 1.0,
                                      "finished_at": 4.0,
                                      "children": [{"name": "foo1",
                                                    "started_at": 2.0,
                                                    "finished_at": 3.0,
                                                    "children": []}]}]},
         "expected": True},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": [], "complete": []},
                  "atomic_actions": [{"name": "foo", "started_at": 1.0,
                                      "finished_at": 4.0,
                                      "children": [{"name": "foo1",
                                                    "started_at": 20,
                                                    "finished_at": 30,
                                                    "children": []}]}]}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": [], "complete": []},
                  "atomic_actions": [{"name": "foo", "started_at": 1.0,
                                      "finished_at": 4.0,
                                      "children": [{"name": "foo1",
                                                    "started_at": 2.0,
                                                    "finished_at": 3.0}]}]}},
        {"data": {"duration": 1, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": [], "complete": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1, "idle_duration": 1.0,
                  "error": [], "output": {"additive": [], "complete": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1,
                  "error": [], "output": {"additive": [], "complete": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": "foo", "output": {"additive": [], "complete": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"complete": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {}, "atomic_actions": []}},
        {"data": {"timestamp": 1.0, "idle_duration": 1.0, "error": [],
                  "output": {"additive": [], "complete": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "idle_duration": 1.0, "error": [],
                  "output": {"additive": [], "complete": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "error": [],
                  "output": {"additive": [], "complete": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "output": {"additive": [], "complete": []},
                  "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "atomic_actions": []}},
        {"data": {"duration": 1.0, "timestamp": 1.0, "idle_duration": 1.0,
                  "error": [], "output": {"additive": [], "complete": []}}},
        {"data": []},
        {"data": {}},
        {"data": "foo"})
    @ddt.unpack
    @mock.patch("rally.common.objects.task.LOG")
    @mock.patch("rally.common.objects.task.charts.validate_output")
    def test_result_has_valid_schema(self, mock_validate_output, mock_log,
                                     data, expected=False,
                                     validate_output_return_value=None,
                                     validate_output_calls=None):
        task = objects.Task(task=self.task)
        mock_validate_output.return_value = validate_output_return_value
        self.assertEqual(expected,
                         task.result_has_valid_schema(data),
                         message=repr(data))
        if validate_output_calls:
            mock_validate_output.assert_has_calls(
                [mock.call(*args) for args in validate_output_calls],
                any_order=True)


class SubtaskTestCase(test.TestCase):

    def setUp(self):
        super(SubtaskTestCase, self).setUp()
        self.subtask = {
            "task_uuid": "00ef46a2-c5b8-4aea-a5ca-0f54a10cbca1",
            "uuid": "00ef46a2-c5b8-4aea-a5ca-0f54a10cbca2",
            "title": "foo",
        }

    @mock.patch("rally.common.objects.task.db.subtask_create")
    def test_init(self, mock_subtask_create):
        mock_subtask_create.return_value = self.subtask
        subtask = objects.Subtask("bar", title="foo")
        mock_subtask_create.assert_called_once_with(
            "bar", title="foo", contexts=None, description=None)
        self.assertEqual(subtask["uuid"], self.subtask["uuid"])

    @mock.patch("rally.common.objects.task.db.subtask_update")
    @mock.patch("rally.common.objects.task.db.subtask_create")
    def test_update_status(self, mock_subtask_create, mock_subtask_update):
        mock_subtask_create.return_value = self.subtask
        subtask = objects.Subtask("bar", title="foo")
        subtask.update_status(consts.SubtaskStatus.FINISHED)
        mock_subtask_update.assert_called_once_with(
            self.subtask["uuid"], {"status": consts.SubtaskStatus.FINISHED})

    @mock.patch("rally.common.objects.task.Workload")
    @mock.patch("rally.common.objects.task.db.subtask_create")
    def test_add_workload(self, mock_subtask_create, mock_workload):
        mock_subtask_create.return_value = self.subtask
        subtask = objects.Subtask("bar", title="foo")

        name = "w"
        description = "descr"
        position = 0
        runner_type = "runner"
        runner = {}
        contexts = {"users": {}}
        sla = {"failure_rate": {"max": 0}}
        args = {"arg": "xxx"}
        hooks = [{"foo": "bar"}]

        workload = subtask.add_workload(
            name, description=description, position=position,
            runner_type=runner_type, runner=runner, contexts=contexts, sla=sla,
            args=args, hooks=hooks)
        mock_workload.assert_called_once_with(
            task_uuid=self.subtask["task_uuid"],
            subtask_uuid=self.subtask["uuid"], name=name,
            description=description, position=position,
            runner_type=runner_type, runner=runner,
            contexts=contexts, sla=sla, args=args,
            hooks=[{"config": h} for h in hooks])
        self.assertIs(workload, mock_workload.return_value)


class WorkloadTestCase(test.TestCase):

    def setUp(self):
        super(WorkloadTestCase, self).setUp()
        self.workload = {
            "task_uuid": "00ef46a2-c5b8-4aea-a5ca-0f54a10cbca1",
            "subtask_uuid": "00ef46a2-c5b8-4aea-a5ca-0f54a10cbca2",
            "uuid": "00ef46a2-c5b8-4aea-a5ca-0f54a10cbca3",
        }

    @mock.patch("rally.common.objects.task.db.workload_create")
    def test_init(self, mock_workload_create):
        mock_workload_create.return_value = self.workload
        name = "w"
        description = "descr"
        position = 0
        runner_type = "constant"
        runner = {"times": 3}
        contexts = {"users": {}}
        sla = {"failure_rate": {"max": 0}}
        args = {"arg": "xxx"}
        hooks = [{"config": {"foo": "bar"}}]
        workload = objects.Workload("uuid1", "uuid2", name=name,
                                    description=description, position=position,
                                    runner=runner, runner_type=runner_type,
                                    contexts=contexts, sla=sla,
                                    args=args, hooks=hooks)
        mock_workload_create.assert_called_once_with(
            task_uuid="uuid1", subtask_uuid="uuid2", name=name, hooks=hooks,
            description=description, position=position, runner=runner,
            runner_type="constant", contexts=contexts, sla=sla, args=args)
        self.assertEqual(workload["uuid"], self.workload["uuid"])

    @mock.patch("rally.common.objects.task.db.workload_data_create")
    @mock.patch("rally.common.objects.task.db.workload_create")
    def test_add_workload_data(self, mock_workload_create,
                               mock_workload_data_create):
        mock_workload_create.return_value = self.workload
        workload = objects.Workload("uuid1", "uuid2", name="w",
                                    description="descr", position=0,
                                    runner_type="foo", runner={},
                                    contexts=None,
                                    sla=None, args=None, hooks=[])

        workload.add_workload_data(0, {"data": "foo"})
        mock_workload_data_create.assert_called_once_with(
            self.workload["task_uuid"], self.workload["uuid"],
            0, {"data": "foo"})

    @mock.patch("rally.common.objects.task.db.workload_set_results")
    @mock.patch("rally.common.objects.task.db.workload_create")
    def test_set_results(self, mock_workload_create,
                         mock_workload_set_results):
        mock_workload_create.return_value = self.workload
        name = "w"
        description = "descr"
        position = 0
        runner_type = "constant"
        runner = {"times": 3}
        contexts = {"users": {}}
        sla = {"failure_rate": {"max": 0}}
        args = {"arg": "xxx"}
        load_duration = 88
        full_duration = 99
        start_time = 1231231277.22
        sla_results = []
        hooks = []
        contexts_results = [{"name": "setup:something"}]
        workload = objects.Workload("uuid1", "uuid2", name=name,
                                    description=description, position=position,
                                    runner=runner, runner_type=runner_type,
                                    contexts=contexts, sla=sla, args=args,
                                    hooks=hooks)

        workload.set_results(load_duration=load_duration,
                             full_duration=full_duration,
                             start_time=start_time,
                             sla_results=sla_results,
                             contexts_results=contexts_results)
        mock_workload_set_results.assert_called_once_with(
            workload_uuid=self.workload["uuid"],
            subtask_uuid=self.workload["subtask_uuid"],
            task_uuid=self.workload["task_uuid"],
            load_duration=load_duration, full_duration=full_duration,
            start_time=start_time, sla_results=sla_results,
            contexts_results=contexts_results,
            hooks_results=None)

    def test_to_task(self):
        workload = {
            "id": 777,
            "uuid": "uuiiidd",
            "task_uuid": "task-uuid",
            "subtask_uuid": "subtask-uuid",
            "name": "Foo.bar",
            "description": "Make something useful (or not).",
            "position": 3,
            "runner_type": "constant",
            "runner": {"times": 3},
            "contexts": {"users": {}},
            "sla": {"failure_rate": {"max": 0}},
            "args": {"key1": "value1"},
            "hooks": [{"config": {
                "action": ["foo", {"arg1": "v1"}],
                "trigger": ["bar", {"arg2": "v2"}]
            }}],
            "sla_results": {"sla": []},
            "context_execution": {},
            "start_time": "2997.23.12",
            "load_duration": 42,
            "full_duration": 37,
            "min_duration": 1,
            "max_duration": 2,
            "total_iteration_count": 7,
            "failed_iteration_count": 2,
            "statistics": {},
            "pass_sla": False
        }
        expected_task = collections.OrderedDict([
            ("version", 2),
            ("title", "A cropped version of a bigger task."),
            ("description", "Auto-generated task from a single workload "
                            "(uuid=%s)" % workload["uuid"]),
            ("subtasks",
             [collections.OrderedDict([
                 ("title", workload["name"]),
                 ("description", workload["description"]),
                 ("scenario", {workload["name"]: workload["args"]}),
                 ("contexts", workload["contexts"]),
                 ("runner", {"constant": {"times": 3}}),
                 ("hooks", [{"action": {"foo": {"arg1": "v1"}},
                             "trigger": {"bar": {"arg2": "v2"}},
                             "description": None}]),
                 ("sla", workload["sla"])])])])
        self.assertEqual(expected_task, objects.Workload.to_task(workload))



# Copyright 2013: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the Test engine."""

import collections
import threading

import mock

from rally.common import objects
from rally import consts
from rally import exceptions
from rally.task import context
from rally.task import engine
from rally.task import scenario
from rally.task import task_cfg
from tests.unit import test


class MyException(exceptions.RallyException):
    msg_fmt = "MyException"


class TaskEngineTestCase(test.TestCase):

    @staticmethod
    def _make_workload(name, args=None, description=None, contexts=None,
                       sla=None, runner=None, hooks=None, position=0):
        return {"uuid": "foo",
                "name": name,
                "position": position,
                "description": description,
                "args": args,
                "contexts": contexts or {},
                "runner_type": runner[0] if runner else "serial",
                "runner": runner[1] if runner else {},
                "sla": sla or {},
                "hooks": hooks or []}

    def test_init(self):
        config = mock.MagicMock()
        task = mock.MagicMock()
        eng = engine.TaskEngine(config, task, mock.Mock())
        self.assertEqual(eng.config, config)
        self.assertEqual(eng.task, task)

    @mock.patch("jsonschema.validate")
    def test_validate(self, mock_validate):
        config = mock.MagicMock()
        eng = engine.TaskEngine(config, mock.MagicMock(),
                                mock.Mock())
        mock_validate = mock.MagicMock()

        eng._validate_config_syntax = mock_validate.syntax
        eng._validate_config_platforms = mock_validate.platforms
        eng._validate_config_semantic = mock_validate.semantic

        eng.validate()

        mock_validate.syntax.assert_called_once_with(config)
        mock_validate.platforms.assert_called_once_with(config)
        mock_validate.semantic.assert_called_once_with(config)

    def test_validate__wrong_syntax(self):
        task = mock.MagicMock()
        eng = engine.TaskEngine(mock.MagicMock(), task, mock.Mock())
        e = exceptions.InvalidTaskConfig(name="foo", pos=0, config="",
                                         reason="foo")
        eng._validate_config_syntax = mock.MagicMock(side_effect=e)
        eng._validate_config_platforms = mock.Mock()

        actual_e = self.assertRaises(exceptions.InvalidTaskException,
                                     eng.validate)
        self.assertEqual(e, actual_e)

        self.assertTrue(task.set_failed.called)
        # the next validation step should not be processed
        self.assertFalse(eng._validate_config_platforms.called)

    def test_validate__wrong_semantic(self):
        task = mock.MagicMock()
        eng = engine.TaskEngine(mock.MagicMock(), task, mock.Mock())
        e = exceptions.InvalidTaskConfig(name="foo", pos=0, config="",
                                         reason="foo")
        eng._validate_config_syntax = mock.MagicMock()
        eng._validate_config_platforms = mock.MagicMock()
        eng._validate_config_semantic = mock.MagicMock(side_effect=e)

        actual_e = self.assertRaises(exceptions.InvalidTaskException,
                                     eng.validate)
        self.assertEqual(e, actual_e)
        self.assertTrue(task.set_failed.called)
        # all steps of validation are called, which means that the last one is
        # failed
        self.assertTrue(eng._validate_config_syntax)
        self.assertTrue(eng._validate_config_platforms)
        self.assertTrue(eng._validate_config_semantic)

    @mock.patch("rally.task.engine.scenario.Scenario.get")
    @mock.patch("rally.task.sla.SLA.validate")
    @mock.patch("rally.task.hook.HookTrigger.validate")
    @mock.patch("rally.task.hook.HookAction.validate")
    @mock.patch("rally.task.engine.runner.ScenarioRunner.validate")
    @mock.patch("rally.task.engine.context.Context.validate")
    def test__validate_workload(
            self, mock_context_validate,
            mock_scenario_runner_validate,
            mock_hook_action_validate,
            mock_hook_trigger_validate,
            mock_sla_validate,
            mock_scenario_get):

        mock_context_validate.return_value = []
        mock_sla_validate.return_value = []
        mock_hook_action_validate.return_value = []
        mock_hook_trigger_validate.return_value = []
        default_context = {"foo": "foo_conf"}
        scenario_cls = mock_scenario_get.return_value
        scenario_cls.get_platform.return_value = "default"
        scenario_cls.get_default_context.return_value = default_context

        scenario_name = "Foo.bar"
        runner_type = "MegaRunner"
        hook_conf = {"action": ("c", "c_args"),
                     "trigger": ("d", "d_args")}
        workload = {"name": scenario_name,
                    "runner_type": runner_type,
                    "runner": {},
                    "contexts": {"a": "a_conf"},
                    "hooks": [hook_conf],
                    "sla": {"foo_sla": "sla_conf"},
                    "position": 2}

        eng = engine.TaskEngine(mock.MagicMock(), mock.MagicMock(),
                                mock.Mock())

        eng._validate_workload(workload)

        mock_scenario_runner_validate.assert_called_once_with(
            name=runner_type, context=None, config=None,
            plugin_cfg={}, vtype=None)
        self.assertEqual([mock.call(name="a",
                                    context=None,
                                    config=None,
                                    plugin_cfg="a_conf",
                                    vtype=None),
                          mock.call(name="foo",
                                    context=None,
                                    config=None,
                                    plugin_cfg="foo_conf",
                                    allow_hidden=True,
                                    vtype=None)],
                         mock_context_validate.call_args_list)
        mock_sla_validate.assert_called_once_with(
            config=None, context=None,
            name="foo_sla", plugin_cfg="sla_conf", vtype=None)
        mock_hook_action_validate.assert_called_once_with(
            config=None, context=None, name="c", plugin_cfg="c_args",
            vtype=None)
        mock_hook_trigger_validate.assert_called_once_with(
            config=None, context=None, name="d", plugin_cfg="d_args",
            vtype=None)

    @mock.patch("rally.task.engine.json.dumps")
    @mock.patch("rally.task.engine.scenario.Scenario.get")
    @mock.patch("rally.task.engine.runner.ScenarioRunner.validate")
    def test___validate_workload__wrong_runner(
            self, mock_scenario_runner_validate,
            mock_scenario_get, mock_dumps):
        mock_dumps.return_value = "<JSON>"
        mock_scenario_runner_validate.return_value = [
            "There is no such runner"]
        scenario_cls = mock_scenario_get.return_value
        scenario_cls.get_default_context.return_value = {}
        workload = self._make_workload(name="sca", runner=("b", {}))
        eng = engine.TaskEngine(mock.MagicMock(), mock.MagicMock(),
                                mock.Mock())

        e = self.assertRaises(exceptions.InvalidTaskConfig,
                              eng._validate_workload, workload)
        self.assertEqual("Input task is invalid!\n\nSubtask sca[0] has wrong "
                         "configuration\nSubtask configuration:\n"
                         "<JSON>\n\nReason(s):\n"
                         " There is no such runner", e.format_message())

    @mock.patch("rally.task.engine.json.dumps")
    @mock.patch("rally.task.engine.scenario.Scenario.get")
    @mock.patch("rally.task.engine.context.Context.validate")
    def test__validate_config_syntax__wrong_context(
            self, mock_context_validate, mock_scenario_get, mock_dumps):
        mock_dumps.return_value = "<JSON>"
        mock_context_validate.return_value = ["context_error"]
        scenario_cls = mock_scenario_get.return_value
        scenario_cls.get_default_context.return_value = {}
        mock_task_instance = mock.MagicMock()
        mock_task_instance.subtasks = [{"workloads": [
            self._make_workload(name="sca"),
            self._make_workload(name="sca", position=1,
                                contexts={"a": "a_conf"})
        ]}]
        eng = engine.TaskEngine(mock.MagicMock(), mock.MagicMock(),
                                mock.Mock())

        e = self.assertRaises(exceptions.InvalidTaskConfig,
                              eng._validate_config_syntax, mock_task_instance)
        self.assertEqual("Input task is invalid!\n\nSubtask sca[1] has wrong "
                         "configuration\nSubtask configuration:\n<JSON>\n\n"
                         "Reason(s):\n context_error", e.format_message())

    @mock.patch("rally.task.engine.json.dumps")
    @mock.patch("rally.task.engine.scenario.Scenario.get")
    @mock.patch("rally.task.sla.SLA.validate")
    def test__validate_config_syntax__wrong_sla(
            self, mock_sla_validate, mock_scenario_get, mock_dumps):
        mock_dumps.return_value = "<JSON>"
        mock_sla_validate.return_value = ["sla_error"]
        scenario_cls = mock_scenario_get.return_value
        scenario_cls.get_default_context.return_value = {}
        mock_task_instance = mock.MagicMock()
        mock_task_instance.subtasks = [{"workloads": [
            self._make_workload(name="sca"),
            self._make_workload(name="sca", position=1,
                                sla={"foo_sla": "sla_conf"})
        ]}]
        eng = engine.TaskEngine(mock.MagicMock(), mock.MagicMock(),
                                mock.Mock())
        e = self.assertRaises(exceptions.InvalidTaskConfig,
                              eng._validate_config_syntax, mock_task_instance)
        self.assertEqual("Input task is invalid!\n\n"
                         "Subtask sca[1] has wrong configuration\n"
                         "Subtask configuration:\n<JSON>\n\n"
                         "Reason(s):\n sla_error", e.format_message())

    @mock.patch("rally.task.engine.json.dumps")
    @mock.patch("rally.task.engine.scenario.Scenario.get")
    @mock.patch("rally.task.hook.HookAction.validate")
    @mock.patch("rally.task.hook.HookTrigger.validate")
    def test__validate_config_syntax__wrong_hook(
            self, mock_hook_trigger_validate,
            mock_hook_action_validate,
            mock_scenario_get, mock_dumps):
        mock_dumps.return_value = "<JSON>"
        mock_hook_trigger_validate.return_value = []
        mock_hook_action_validate.return_value = ["hook_error"]
        scenario_cls = mock_scenario_get.return_value
        scenario_cls.get_default_context.return_value = {}
        mock_task_instance = mock.MagicMock()
        hook_conf = {"action": ("c", "c_args"),
                     "trigger": ("d", "d_args")}
        mock_task_instance.subtasks = [{"workloads": [
            self._make_workload(name="sca"),
            self._make_workload(name="sca", position=1,
                                hooks=[hook_conf])
        ]}]
        eng = engine.TaskEngine(mock.MagicMock(), mock.MagicMock(),
                                mock.Mock())

        e = self.assertRaises(exceptions.InvalidTaskConfig,
                              eng._validate_config_syntax, mock_task_instance)

        self.assertEqual("Input task is invalid!\n\n"
                         "Subtask sca[1] has wrong configuration\n"
                         "Subtask configuration:\n<JSON>\n\n"
                         "Reason(s):\n hook_error", e.format_message())

    @mock.patch("rally.task.engine.json.dumps")
    @mock.patch("rally.task.engine.scenario.Scenario.get")
    @mock.patch("rally.task.hook.HookTrigger.validate")
    @mock.patch("rally.task.hook.HookAction.validate")
    def test__validate_config_syntax__wrong_trigger(
            self, mock_hook_action_validate,
            mock_hook_trigger_validate,
            mock_scenario_get, mock_dumps):
        mock_dumps.return_value = "<JSON>"
        mock_hook_trigger_validate.return_value = ["trigger_error"]
        mock_hook_action_validate.return_value = []
        scenario_cls = mock_scenario_get.return_value
        scenario_cls.get_default_context.return_value = {}
        mock_task_instance = mock.MagicMock()
        hook_conf = {"action": ("c", "c_args"),
                     "trigger": ("d", "d_args")}
        mock_task_instance.subtasks = [{"workloads": [
            self._make_workload(name="sca"),
            self._make_workload(name="sca", position=1,
                                hooks=[hook_conf])
        ]}]
        eng = engine.TaskEngine(mock.MagicMock(), mock.MagicMock(),
                                mock.Mock())

        e = self.assertRaises(exceptions.InvalidTaskConfig,
                              eng._validate_config_syntax, mock_task_instance)

        self.assertEqual("Input task is invalid!\n\n"
                         "Subtask sca[1] has wrong configuration\n"
                         "Subtask configuration:\n<JSON>\n\n"
                         "Reason(s):\n trigger_error", e.format_message())

    @mock.patch("rally.task.engine.context.ContextManager.cleanup")
    @mock.patch("rally.task.engine.context.ContextManager.setup")
    def test__validate_config_semantic(self, mock_context_manager_setup,
                                       mock_context_manager_cleanup):
        env = mock.MagicMock(uuid="env_uuid")
        env.check_health.return_value = {
            "foo": {"available": True, "message": ""},
            "bar": {"available": True, "message": ""}
        }

        @scenario.configure("SomeScen.scenario")
        class SomeScen(scenario.Scenario):

            def run(self):
                pass

        mock_task_instance = mock.MagicMock()
        wconf1 = self._make_workload(name="SomeScen.scenario")
        wconf2 = self._make_workload(name="SomeScen.scenario",
                                     position=1)
        subtask1 = {"workloads": [wconf1, wconf2]}

        wconf3 = self._make_workload(name="SomeScen.scenario",
                                     position=2)
        subtask2 = {"workloads": [wconf3]}

        mock_task_instance.subtasks = [subtask1, subtask2]
        fake_task = mock.MagicMock()
        eng = engine.TaskEngine(mock_task_instance, fake_task, env)

        eng._validate_config_semantic(mock_task_instance)

        env.check_health.return_value = {
            "foo": {"available": True, "message": ""},
            "bar": {"available": False, "message": "", "traceback": "AAAA"}
        }
        self.assertRaises(exceptions.ValidationError,
                          eng._validate_config_semantic,
                          mock_task_instance)

    @mock.patch("rally.task.engine.TaskEngine._validate_workload")
    def test__validate_config_platforms(self, mock__validate_workload):

        foo_cred = {"admin": "admin", "users": ["user1"]}
        env = mock.MagicMock(data={
            "platforms": {
                "foo": {
                    "platform_name": "foo", "platform_data": foo_cred
                }
            }
        })

        workload1 = "workload1"
        workload2 = "workload2"
        subtasks = [{"workloads": [workload1]},
                    {"workloads": [workload2]}]
        config = mock.Mock(subtasks=subtasks)
        eng = engine.TaskEngine({}, mock.MagicMock(), env)

        eng._validate_config_platforms(config)

        self.assertEqual(
            [mock.call(w, vtype="platform",
                       vcontext={"platforms": {"foo": foo_cred},
                                 "task": eng.task})
             for w in (workload1, workload2)],
            mock__validate_workload.call_args_list)

    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.ResultConsumer")
    @mock.patch("rally.task.engine.context.ContextManager.cleanup")
    @mock.patch("rally.task.engine.context.ContextManager.setup")
    @mock.patch("rally.task.engine.scenario.Scenario")
    @mock.patch("rally.task.engine.runner.ScenarioRunner")
    def test_run__update_status(
            self, mock_scenario_runner, mock_scenario,
            mock_context_manager_setup, mock_context_manager_cleanup,
            mock_result_consumer, mock_task_get_status):

        task = mock.MagicMock()
        mock_task_get_status.return_value = consts.TaskStatus.ABORTING
        eng = engine.TaskEngine(mock.MagicMock(), task, mock.Mock())
        eng.run()
        task.update_status.assert_has_calls([
            mock.call(consts.TaskStatus.RUNNING),
            mock.call(consts.TaskStatus.FINISHED)
        ])

    @mock.patch("rally.task.engine.objects.task.Task.get_status")
    @mock.patch("rally.task.engine.LOG")
    @mock.patch("rally.task.engine.ResultConsumer")
    @mock.patch("rally.task.engine.context.Context")
    @mock.patch("rally.task.engine.scenario.Scenario")
    @mock.patch("rally.task.engine.runner.ScenarioRunner")
    @mock.patch("rally.task.engine.context.ContextManager.cleanup")
    @mock.patch("rally.task.engine.context.ContextManager.setup")
    def test_run_exception_is_logged(
            self, mock_context_manager_setup, mock_context_manager_cleanup,
            mock_scenario_runner, mock_scenario, mock_context,
            mock_result_consumer, mock_log, mock_task_get_status):
        scenario_cls = mock_scenario.get.return_value
        scenario_cls.get_default_context.return_value = {}

        context_cls = mock_context.get.return_value
        context_cls.get_fullname.return_value = "context_a"

        mock_context_manager_setup.side_effect = Exception
        mock_result_consumer.is_task_in_aborting_status.return_value = False

        mock_task_instance = mock.MagicMock()

        mock_task_instance.subtasks = [{
            "title": "foo",
            "description": "Do not launch it!!",
            "contexts": {},
            "workloads": [
                self._make_workload(name="a.task", description="foo",
                                    contexts={"context_a": {"a": 1}}),
                self._make_workload(name="a.task", description="foo",
                                    contexts={"context_a": {"b": 2}},
                                    position=2)]}]
        eng = engine.TaskEngine(mock_task_instance, mock.MagicMock(),
                                mock.MagicMock())

        eng.run()

        self.assertEqual(2, mock_log.exception.call_count)

    @mock.patch("rally.task.engine.ResultConsumer")
    @mock.patch("rally.task.engine.context.ContextManager.cleanup")
    @mock.patch("rally.task.engine.context.ContextManager.setup")
    @mock.patch("rally.task.engine.scenario.Scenario")
    @mock.patch("rally.task.engine.runner.ScenarioRunner")
    def test_run__task_soft_aborted(
            self, mock_scenario_runner, mock_scenario,
            mock_context_manager_setup, mock_context_manager_cleanup,
            mock_result_consumer):
        scenario_cls = mock_scenario.get.return_value
        scenario_cls.get_platform.return_value = "openstack"
        scenario_cls.get_info.return_value = {"title": ""}
        task = mock.MagicMock()
        mock_result_consumer.is_task_in_aborting_status.side_effect = [False,
                                                                       False,
                                                                       True]
        config = task_cfg.TaskConfig({
            "a.task": [{"runner": {"type": "a", "b": 1},
                        "description": "foo"}],
            "b.task": [{"runner": {"type": "a", "b": 1},
                        "description": "bar"}],
            "c.task": [{"runner": {"type": "a", "b": 1},
                        "description": "xxx"}]
        })
        fake_runner_cls = mock.MagicMock()
        fake_runner = mock.MagicMock()
        fake_runner_cls.return_value = fake_runner
        mock_scenario_runner.get.return_value = fake_runner_cls
        eng = engine.TaskEngine(config, task, mock.MagicMock())

        eng.run()

        self.assertEqual(2, fake_runner.run.call_count)
        self.assertEqual(mock.call(consts.TaskStatus.ABORTED),
                         task.update_status.mock_calls[-1])
        subtask_obj = task.add_subtask.return_value
        subtask_obj.update_status.assert_has_calls((
            mock.call(consts.SubtaskStatus.FINISHED),
            mock.call(consts.SubtaskStatus.FINISHED),
            mock.call(consts.SubtaskStatus.ABORTED),
        ))

    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.ResultConsumer")
    @mock.patch("rally.task.engine.context.ContextManager.cleanup")
    @mock.patch("rally.task.engine.context.ContextManager.setup")
    @mock.patch("rally.task.engine.scenario.Scenario")
    @mock.patch("rally.task.engine.runner.ScenarioRunner")
    def test_run__task_aborted(
            self, mock_scenario_runner, mock_scenario,
            mock_context_manager_setup, mock_context_manager_cleanup,
            mock_result_consumer, mock_task_get_status):
        task = mock.MagicMock(spec=objects.Task)
        config = task_cfg.TaskConfig({
            "a.task": [{"runner": {"type": "a", "b": 1}}],
            "b.task": [{"runner": {"type": "a", "b": 1}}],
            "c.task": [{"runner": {"type": "a", "b": 1}}]
        })
        fake_runner_cls = mock.MagicMock()
        fake_runner = mock.MagicMock()
        fake_runner_cls.return_value = fake_runner
        mock_task_get_status.return_value = consts.TaskStatus.SOFT_ABORTING
        mock_scenario_runner.get.return_value = fake_runner_cls
        eng = engine.TaskEngine(config, task, mock.Mock())
        eng.run()
        self.assertEqual(mock.call(consts.TaskStatus.ABORTED),
                         task.update_status.mock_calls[-1])
        subtask_obj = task.add_subtask.return_value
        subtask_obj.update_status.assert_called_once_with(
            consts.SubtaskStatus.ABORTED)

    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.ResultConsumer")
    @mock.patch("rally.task.engine.context.ContextManager.cleanup")
    @mock.patch("rally.task.engine.context.ContextManager.setup")
    @mock.patch("rally.task.engine.scenario.Scenario")
    @mock.patch("rally.task.engine.runner.ScenarioRunner")
    def test_run__subtask_crashed(
            self, mock_scenario_runner, mock_scenario,
            mock_context_manager_setup, mock_context_manager_cleanup,
            mock_result_consumer, mock_task_get_status):
        task = mock.MagicMock(spec=objects.Task)
        subtask_obj = task.add_subtask.return_value
        subtask_obj.add_workload.side_effect = MyException()
        mock_result_consumer.is_task_in_aborting_status.return_value = False
        config = task_cfg.TaskConfig({
            "a.task": [{"runner": {"type": "a", "b": 1}}],
            "b.task": [{"runner": {"type": "a", "b": 1}}],
            "c.task": [{"runner": {"type": "a", "b": 1}}]
        })
        fake_runner_cls = mock.MagicMock()
        fake_runner = mock.MagicMock()
        fake_runner_cls.return_value = fake_runner
        mock_scenario_runner.get.return_value = fake_runner_cls
        eng = engine.TaskEngine(config, task, mock.Mock())
        self.assertRaises(MyException, eng.run)

        task.update_status.assert_has_calls((
            mock.call(consts.TaskStatus.RUNNING),
            mock.call(consts.TaskStatus.CRASHED),
        ))
        subtask_obj.update_status.assert_called_once_with(
            consts.SubtaskStatus.CRASHED)

    def test__prepare_context(self):

        @context.configure("test1", 1, platform="testing")
        class TestContext1(context.Context):
            pass

        self.addCleanup(TestContext1.unregister)

        @context.configure("test2", 2, platform="testing")
        class TestContext2(context.Context):
            pass

        self.addCleanup(TestContext2.unregister)

        @scenario.configure("test_ctx.test", platform="testing",
                            context={"test1@testing": {"a": 1}})
        class TestScenario(scenario.Scenario):
            pass

        self.addCleanup(TestScenario.unregister)

        task = mock.MagicMock()
        name = "test_ctx.test"
        context_config = {"test1": 1, "test2": 2}

        env = mock.MagicMock()
        eng = engine.TaskEngine({}, task, env)

        result = eng._prepare_context(context_config, name, "foo_uuid")

        expected_result = {
            "task": task,
            "owner_id": "foo_uuid",
            "scenario_name": name,
            "config": {"test1@testing": 1, "test2@testing": 2},
            "env": env.data
        }
        self.assertEqual(expected_result, result)


class ResultConsumerTestCase(test.TestCase):

    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.ResultConsumer.wait_and_abort")
    @mock.patch("rally.task.sla.SLAChecker")
    def test_consume_results(
            self, mock_sla_checker, mock_result_consumer_wait_and_abort,
            mock_task_get_status):
        mock_sla_instance = mock.MagicMock()
        mock_sla_checker.return_value = mock_sla_instance
        mock_task_get_status.return_value = consts.TaskStatus.RUNNING
        workload_cfg = {"fake": 2, "hooks": []}
        task = mock.MagicMock()
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)
        runner = mock.MagicMock()

        results = [
            [{"duration": 1, "timestamp": 3}],
            [{"duration": 2, "timestamp": 2}]
        ]

        runner.result_queue = collections.deque(results)
        runner.event_queue = collections.deque()
        ctx_manager = mock.MagicMock()

        with engine.ResultConsumer(workload_cfg, task=task, subtask=subtask,
                                   workload=workload, runner=runner,
                                   abort_on_sla_failure=False,
                                   ctx_manager=ctx_manager) as consumer_obj:
            pass

        mock_sla_instance.add_iteration.assert_has_calls([
            mock.call({"duration": 1, "timestamp": 3}),
            mock.call({"duration": 2, "timestamp": 2})])

        self.assertEqual([{"duration": 2, "timestamp": 2},
                          {"duration": 1, "timestamp": 3}],
                         consumer_obj.results)

    @mock.patch("rally.task.hook.HookExecutor")
    @mock.patch("rally.task.engine.LOG")
    @mock.patch("rally.task.engine.time.time")
    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.ResultConsumer.wait_and_abort")
    @mock.patch("rally.task.sla.SLAChecker")
    def test_consume_results_no_iteration(
            self, mock_sla_checker, mock_result_consumer_wait_and_abort,
            mock_task_get_status, mock_time, mock_log, mock_hook_executor):
        mock_time.side_effect = [0, 1]
        mock_sla_instance = mock.MagicMock()
        mock_sla_results = mock.MagicMock()
        mock_sla_checker.return_value = mock_sla_instance
        mock_sla_instance.results.return_value = mock_sla_results
        mock_task_get_status.return_value = consts.TaskStatus.RUNNING
        workload_cfg = {"fake": 2, "hooks": []}
        task = mock.MagicMock()
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)
        runner = mock.MagicMock()

        results = []
        runner.result_queue = collections.deque(results)
        runner.event_queue = collections.deque()
        ctx_manager = mock.MagicMock()

        with engine.ResultConsumer(workload_cfg, task=task, subtask=subtask,
                                   workload=workload, runner=runner,
                                   abort_on_sla_failure=False,
                                   ctx_manager=ctx_manager):
            pass

        self.assertFalse(workload.add_workload_data.called)
        workload.set_results.assert_called_once_with(
            full_duration=1, sla_results=mock_sla_results, load_duration=0,
            start_time=None,
            contexts_results=ctx_manager.contexts_results())

    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.ResultConsumer.wait_and_abort")
    @mock.patch("rally.task.sla.SLAChecker")
    def test_consume_results_sla_failure_abort(
            self, mock_sla_checker, mock_result_consumer_wait_and_abort,
            mock_task_get_status):
        mock_sla_instance = mock.MagicMock()
        mock_sla_checker.return_value = mock_sla_instance
        mock_sla_instance.add_iteration.side_effect = [True, True, False,
                                                       False]
        workload_cfg = {"fake": 2, "hooks": []}
        task = mock.MagicMock()
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)
        runner = mock.MagicMock()

        runner.result_queue = collections.deque(
            [[{"duration": 1, "timestamp": 1},
              {"duration": 2, "timestamp": 2}]] * 4)
        ctx_manager = mock.MagicMock()

        with engine.ResultConsumer(workload_cfg, task=task, subtask=subtask,
                                   workload=workload, runner=runner,
                                   abort_on_sla_failure=True,
                                   ctx_manager=ctx_manager):
            pass

        self.assertTrue(runner.abort.called)
        task.update_status.assert_called_once_with(
            consts.TaskStatus.SOFT_ABORTING)

    @mock.patch("rally.task.hook.HookExecutor")
    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.threading.Thread")
    @mock.patch("rally.task.engine.threading.Event")
    @mock.patch("rally.task.sla.SLAChecker")
    def test_consume_results_abort_manually(self, mock_sla_checker,
                                            mock_event, mock_thread,
                                            mock_task_get_status,
                                            mock_hook_executor):
        runner = mock.MagicMock(result_queue=False)

        is_done = mock.MagicMock()
        is_done.isSet.side_effect = (False, True)

        task = mock.MagicMock()
        mock_task_get_status.return_value = consts.TaskStatus.ABORTED
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)

        workload_cfg = {"fake": 2, "hooks": []}

        mock_hook_executor_instance = mock_hook_executor.return_value
        ctx_manager = mock.MagicMock()

        with engine.ResultConsumer(workload_cfg, task=task, subtask=subtask,
                                   workload=workload, runner=runner,
                                   abort_on_sla_failure=True,
                                   ctx_manager=ctx_manager):
            pass

        mock_sla_checker.assert_called_once_with(workload_cfg)
        mock_hook_executor.assert_called_once_with(workload_cfg, task)
        self.assertFalse(mock_hook_executor_instance.on_iteration.called)
        mocked_set_aborted = mock_sla_checker.return_value.set_aborted_manually
        mocked_set_aborted.assert_called_once_with()

    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.sla.SLAChecker")
    def test_consume_results_sla_failure_continue(self, mock_sla_checker,
                                                  mock_task_get_status):
        mock_sla_instance = mock.MagicMock()
        mock_sla_checker.return_value = mock_sla_instance
        mock_task_get_status.return_value = consts.TaskStatus.CRASHED
        mock_sla_instance.add_iteration.side_effect = [True, True, False,
                                                       False]
        workload_cfg = {"fake": 2, "hooks": []}
        task = mock.MagicMock()
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)
        runner = mock.MagicMock()
        runner.result_queue = collections.deque(
            [[{"duration": 1, "timestamp": 4}]] * 4)
        runner.event_queue = collections.deque()
        ctx_manager = mock.MagicMock()

        with engine.ResultConsumer(workload_cfg, task=task, subtask=subtask,
                                   workload=workload, runner=runner,
                                   ctx_manager=ctx_manager,
                                   abort_on_sla_failure=False):
            pass

        self.assertEqual(0, runner.abort.call_count)

    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.threading.Thread")
    @mock.patch("rally.task.engine.threading.Event")
    @mock.patch("rally.task.sla.SLAChecker")
    def test_consume_results_with_unexpected_failure(self, mock_sla_checker,
                                                     mock_event, mock_thread,
                                                     mock_task_get_status):
        mock_sla_instance = mock.MagicMock()
        mock_sla_checker.return_value = mock_sla_instance
        workload_cfg = {"fake": 2, "hooks": []}
        task = mock.MagicMock()
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)
        runner = mock.MagicMock()
        runner.result_queue = collections.deque([1])
        runner.event_queue = collections.deque()
        ctx_manager = mock.MagicMock()
        exc = MyException()
        try:
            with engine.ResultConsumer(workload_cfg, task=task,
                                       subtask=subtask, workload=workload,
                                       runner=runner, ctx_manager=ctx_manager,
                                       abort_on_sla_failure=False):
                raise exc
        except MyException:
            pass
        else:
            self.fail("ResultConsumer should re-raise the exception.")

        mock_sla_instance.set_unexpected_failure.assert_has_calls(
            [mock.call(exc)])

    @mock.patch("rally.task.engine.CONF")
    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.ResultConsumer.wait_and_abort")
    @mock.patch("rally.task.sla.SLAChecker")
    def test_consume_results_chunked(
            self, mock_sla_checker, mock_result_consumer_wait_and_abort,
            mock_task_get_status, mock_conf):
        mock_conf.raw_result_chunk_size = 2
        mock_sla_instance = mock.MagicMock()
        mock_sla_checker.return_value = mock_sla_instance
        mock_task_get_status.return_value = consts.TaskStatus.RUNNING
        workload_cfg = {"fake": 2, "hooks": []}
        task = mock.MagicMock(spec=objects.Task)
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)
        runner = mock.MagicMock()

        results = [
            [{"duration": 1, "timestamp": 3},
             {"duration": 2, "timestamp": 2},
             {"duration": 3, "timestamp": 3}],
            [{"duration": 4, "timestamp": 2},
             {"duration": 5, "timestamp": 3}],
            [{"duration": 6, "timestamp": 2}],
            [{"duration": 7, "timestamp": 1}],
        ]

        runner.result_queue = collections.deque(results)
        runner.event_queue = collections.deque()
        ctx_manager = mock.MagicMock()

        with engine.ResultConsumer(workload_cfg, task=task, subtask=subtask,
                                   workload=workload, runner=runner,
                                   abort_on_sla_failure=False,
                                   ctx_manager=ctx_manager) as consumer_obj:
            pass

        mock_sla_instance.add_iteration.assert_has_calls([
            mock.call({"duration": 1, "timestamp": 3}),
            mock.call({"duration": 2, "timestamp": 2}),
            mock.call({"duration": 3, "timestamp": 3}),
            mock.call({"duration": 4, "timestamp": 2}),
            mock.call({"duration": 5, "timestamp": 3}),
            mock.call({"duration": 6, "timestamp": 2}),
            mock.call({"duration": 7, "timestamp": 1})])

        self.assertEqual([{"duration": 7, "timestamp": 1}],
                         consumer_obj.results)

        workload.add_workload_data.assert_has_calls([
            mock.call(0, {"raw": [{"duration": 2, "timestamp": 2},
                                  {"duration": 1, "timestamp": 3}]}),
            mock.call(1, {"raw": [{"duration": 4, "timestamp": 2},
                                  {"duration": 3, "timestamp": 3}]}),
            mock.call(2, {"raw": [{"duration": 6, "timestamp": 2},
                                  {"duration": 5, "timestamp": 3}]}),
            mock.call(3, {"raw": [{"duration": 7, "timestamp": 1}]})])

    @mock.patch("rally.task.engine.LOG")
    @mock.patch("rally.task.hook.HookExecutor")
    @mock.patch("rally.task.engine.time.time")
    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.ResultConsumer.wait_and_abort")
    @mock.patch("rally.task.sla.SLAChecker")
    def test_consume_events(
            self, mock_sla_checker, mock_result_consumer_wait_and_abort,
            mock_task_get_status, mock_time, mock_hook_executor, mock_log):
        mock_time.side_effect = [0, 1]
        mock_sla_instance = mock_sla_checker.return_value
        mock_sla_results = mock_sla_instance.results.return_value
        mock_hook_executor_instance = mock_hook_executor.return_value
        mock_hook_results = mock_hook_executor_instance.results.return_value

        mock_task_get_status.return_value = consts.TaskStatus.RUNNING
        workload_cfg = {"fake": 2, "hooks": [{"config": True}]}
        task = mock.MagicMock()
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)
        runner = mock.MagicMock()
        events = [
            {"type": "iteration", "value": 1},
            {"type": "iteration", "value": 2},
            {"type": "iteration", "value": 3}
        ]
        runner.result_queue = collections.deque()
        runner.event_queue = collections.deque(events)

        ctx_manager = mock.MagicMock()
        consumer_obj = engine.ResultConsumer(
            workload_cfg, task=task, subtask=subtask, workload=workload,
            runner=runner, abort_on_sla_failure=False, ctx_manager=ctx_manager)
        stop_event = threading.Event()

        def set_stop_event(event_type, value):
            if not runner.event_queue:
                stop_event.set()

        mock_hook_executor_instance.on_event.side_effect = set_stop_event

        with consumer_obj:
            stop_event.wait(1)

        mock_hook_executor_instance.on_event.assert_has_calls([
            mock.call(event_type="iteration", value=1),
            mock.call(event_type="iteration", value=2),
            mock.call(event_type="iteration", value=3)
        ])

        self.assertFalse(workload.add_workload_data.called)
        workload.set_results.assert_called_once_with(
            full_duration=1,
            load_duration=0,
            sla_results=mock_sla_results,
            hooks_results=mock_hook_results,
            start_time=None,
            contexts_results=ctx_manager.contexts_results())

    @mock.patch("rally.task.engine.threading.Thread")
    @mock.patch("rally.task.engine.threading.Event")
    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.TaskEngine._prepare_context")
    @mock.patch("rally.task.engine.time.sleep")
    def test_wait_and_abort_on_abort(
            self, mock_sleep, mock_task_engine__prepare_context,
            mock_task_get_status, mock_event, mock_thread):
        runner = mock.MagicMock()
        workload_cfg = mock.MagicMock()
        task = mock.MagicMock()
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)
        mock_task_get_status.side_effect = (consts.TaskStatus.RUNNING,
                                            consts.TaskStatus.RUNNING,
                                            consts.TaskStatus.ABORTING)
        mock_is_done = mock.MagicMock()
        mock_event.return_value = mock_is_done
        mock_is_done.isSet.return_value = False
        ctx_manager = mock.MagicMock()

        res = engine.ResultConsumer(workload_cfg, task=task, subtask=subtask,
                                    workload=workload, runner=runner,
                                    abort_on_sla_failure=True,
                                    ctx_manager=ctx_manager)
        res.wait_and_abort()

        runner.abort.assert_called_with()
        # test task.get_status is checked until is_done is not set
        self.assertEqual(3, mock_task_get_status.call_count)

    @mock.patch("rally.task.engine.threading.Thread")
    @mock.patch("rally.task.engine.threading.Event")
    @mock.patch("rally.common.objects.Task.get_status")
    @mock.patch("rally.task.engine.TaskEngine._prepare_context")
    @mock.patch("rally.task.engine.time.sleep")
    def test_wait_and_abort_on_no_abort(
            self, mock_sleep, mock_task_engine__prepare_context,
            mock_task_get_status, mock_event, mock_thread):
        runner = mock.MagicMock()
        workload_cfg = mock.MagicMock()
        task = mock.MagicMock()
        subtask = mock.Mock(spec=objects.Subtask)
        workload = mock.Mock(spec=objects.Workload)
        mock_task_get_status.return_value = consts.TaskStatus.RUNNING
        mock_is_done = mock.MagicMock()
        mock_event.return_value = mock_is_done
        ctx_manager = mock.MagicMock()

        mock_is_done.isSet.side_effect = [False, False, False, False, True]

        res = engine.ResultConsumer(workload_cfg, task=task, subtask=subtask,
                                    workload=workload, runner=runner,
                                    abort_on_sla_failure=True,
                                    ctx_manager=ctx_manager)
        res.wait_and_abort()

        # check method don't abort runner if task is not aborted
        self.assertFalse(runner.abort.called)
        # test task.get_status is checked until is_done is not set
        self.assertEqual(4, mock_task_get_status.call_count)
